﻿# coding = utf-8
###################################################################################
#              Toolset for quick assessment of potential locations                #
#                     for small water reservoirs placement                        #
#                        Created by Jan Devátý, 2017                              #
#          Department of Irrigation, Drainage and Landscape Engineering           #
#            Faculty of Civil Engineering, Czech Technical University             #
#                                                                                 #
###################################################################################
import sys

import arcpy
import os
import locale
import math
import datetime
#from osgeo import ogr

locale.setlocale(locale.LC_ALL, '')
sr = arcpy.SpatialReference(5514)
arcpy.env.overwriteOutput = True

## FILE NAMES ##
taskSettingsFileName = "_settings"
profilesFileName = "profiles"
profileElevFileName = "profiles_elev"
profilesPropsTableName = "profiles_props"
profileMinPointsFileName = "min_points"
insidePointsFileName = "inside_points"
charLinesFileName = "char_lines_"
waterLinesFileName = "water_lines"

# generated contours
prefixCont = "cont_"
# intersection of a contour and a dam profile
prefixContPoly = "cont_poly_"
# selected polygons of contour-dam intersection
prefixPoly = "flooded_area_"
# statistics table for each level polygon
prefixStatsTable = "stats_"
# final profile properties table name prefix
prefixPropsTable = "profile_"
# dam line for given profile and flooding level
damAxisPrefix = "dam_"
prefixMaxCont = "max_contour_"
prefixMaxContPoly = "max_cont_poly_"
# individual DEMs
DEMfileNamePrefix = "DEM_"
DEMwDamFileNamePrefix = "DEMwDam_"
damPointsNamePrefix = "dam_points_"
damDEMFileNamePrefix = "DamDEM_"
damPlanPrefix = "dam_plan_"
# level-volume curve tables
prefixVolumes = "volumes_"

## FIELD NAMES ##
IDfieldName = "IDFieldName"
levelIntervalFieldName = "interval"
startElevFieldFieldName = "startElevField"
taskDEMFieldName = "DEM"
damCrestWidthFieldName = "crestWidth"
damSaveHeightFieldName = "heightAWL"
damWatersideSlopeFieldName = "waterSlope"
damWindwardSlopeFieldName = "windSlope"
damFoundationDepthFieldName = "fDepth"
shapeFieldName = "SHAPE@"
minElevFieldName = "minElev"
maxLeftFieldName = "maxLeft"
maxRightFieldName = "maxRight"
rangeFieldName = "elevRange"
startElevFieldName = "startElev"
minLevelFieldName = "minLevel"
maxLevelFieldName = "maxLevel"
watershedAreaFieldName = "wsArea"
levelFieldName = "level"
damHeightFieldName = "damHeight"
maxDepthFieldName = "maxDepth"
deepestPntElevFieldName = "deepestPntElev"
wLevelIDFieldName = "wlID"
elevationFieldName = "WLevel"
sideFieldName = "side"
meanDepthFieldName = "meanDepth"
floodedAreaFieldName = "flArea"
volumeFieldName = "volume"
damLengthFieldName = "damLength"
damVolumeFieldName = "damVolume"
damFVolumeFieldName = "fdamVolume"
volumeCoefFieldName = "volumeCoef"
urbanConfFieldName = "xUrban"
roadsConfFieldName = "xRoads"
sRoadsConfFieldName = "xServiceR"
railConfFieldName = "xRails"

# FUNCTIONS DEFINITION ------------------------------------------------------------------------------------------
def deleteTemporaryFiles(toBeDeleted):
    if len(toBeDeleted)>0:
        arcpy.AddMessage("Deleting temporary files ... ")
        #arcpy.AddMessage(str(toBeDeleted))
        for file in toBeDeleted:
            try:
                arcpy.management.Delete(file, "")
                pass
            except Exception as e:
                arcpy.AddError(e.message)
                arcpy.AddError("Deleting temporary file '"+str(file)+"' failed!")        
        arcpy.AddMessage(" ... successfull.")

def getFieldNames(featureClass):
    fieldNames = []
    fields = arcpy.ListFields(featureClass)
    for field in fields:
        fieldNames.append(field.name)
    return fieldNames    

# checks if file extension is correct for given workspace type, returns correct file path
def checkFileExtension(testPath, ext):
    dirname = os.path.dirname(testPath)
    rightPath = testPath
    if arcpy.Describe(dirname).dataType == "Folder":
        filename = os.path.basename(testPath)
        filext = filename.split(".")
        if len(filext) == 1:
            rightPath = testPath+"."+ext
        if len(filext) > 1:
            extension = filext[-1]
            if extension != ext:
                dirname = os.path.dirname(testPath)
                filename = ".".join(filext[:-1])
                rightPath = os.path.join(dirname, filename)+"."+ext
    return rightPath

class Toolbox(object):
    def __init__(self):
        """Define the toolbox (the name of the toolbox is the name of the .pyt file)."""
        self.label = "MVNtool - potencial dam profile analysis"
        self.alias = ""

        # List of tool classes associated with this toolbox
        self.tools = [ProcessProfiles, FindConflicts, AddResultsToMap, CopyProperties, BreachDams, CreateSummary, CalculateLevelVolumeCurve, GetDamsWatershed]

class ProcessProfiles(object):

    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "1. Process profiles"
        self.description = "Prepares dam profiles and DEM with dams included"
        self.canRunInBackground = False
        self.category = ""
        #locale.setlocale(locale.LC_ALL, "csy")

    def getParameterInfo(self):
        """Define parameter definitions"""
        inputProfiles = arcpy.Parameter(
            displayName = "Dam profiles feature layer",
            name = "inputProfiles",
            datatype = "GPFeatureLayer",
            parameterType = "Required",
            direction = "Input"
            )
        profileIDfieldName = arcpy.Parameter(
            displayName = "Attribute field with the profiles identifier",
            name = "profileIDfieldName",
            datatype = "Field",
            parameterType = "Required",
            direction = "Input"
            )
        profileIDfieldName.parameterDependencies = [inputProfiles.name]
        
        inputDEM = arcpy.Parameter(
            displayName = "Input DEM raster layer",
            name = "inputDEM",
            datatype = "GPRasterLayer",
            parameterType = "Required",
            direction = "Input"
            )
        targetWorkspace = arcpy.Parameter(
            displayName = "Task geodatabase",
            name = "targetWorkspace",
            datatype = "DEWorkspace",
            parameterType = "Required",
            direction = "Input"
            )
        saveProfiles = arcpy.Parameter(
            displayName = "Save profiles cross-sections into .csv files",
            name = "saveProfiles",
            datatype = "GPBoolean",
            parameterType = "Optional",
            direction = "Input"
            )
        saveProfiles.value = False
        
        profilesOutputTable = arcpy.Parameter(
            displayName = "File to store the tabelated cross-sections of profiles",
            name = "profilesOutputTable",
            datatype = "DEFile",
            parameterType = "Optional",
            direction = "Output"
            )
        
        depthStep = arcpy.Parameter(
            displayName = "Elevation step to create the water levels [m]",
            name = "depthStep",
            datatype = "GPDouble",
            parameterType = "Required",
            direction = "Input",
            category = "Task settings"
            )
        depthStep.value = 1
        
        maxDamHeigth = arcpy.Parameter(
            displayName = "Maximum potencial height of the dam [m]",
            name = "maxDamHeigth",
            datatype = "GPDouble",
            parameterType = "Required",
            direction = "Input",
            category = "Task settings"
            )
        maxDamHeigth.value = 9

        useMinElevation = arcpy.Parameter(
            displayName = "Use minimum elevation of first water level from profiles attribute field",
            name = "useMinElevation",
            datatype = "GPBoolean",
            parameterType = "Optional",
            direction = "Input",
            category = "Task settings")
        useMinElevation.value = False
        
        minElevationField = arcpy.Parameter(
            displayName = "The minimum elevation value field",
            name = "minElevationField",
            datatype = "Field",
            parameterType = "Optional",
            direction = "Input",
            category = "Task settings"
            )
        minElevationField.parameterDependencies = [inputProfiles.name]
        minElevationField.enabled = False

        eraseIntermediateData = arcpy.Parameter(
            displayName = "Erase the intermediate datasets used for the processing",
            name = "eraseIntermediate",
            datatype = "GPBoolean",
            parameterType = "Required",
            direction = "Input",
            category = "Task settings")
        eraseIntermediateData.value = True

        damCrestWidth = arcpy.Parameter(
            displayName = "Dam crest width [m]",
            name = "damCrestWidth",
            datatype = "GPDouble",
            parameterType = "Required",
            direction = "Input",
            category = "Dam settings"
            )
        damCrestWidth.value = 3
        
        damSaveHeight = arcpy.Parameter(
            displayName = "Height above max water level [m]",
            name = "damSaveHeight",
            datatype = "GPDouble",
            parameterType = "Required",
            direction = "Input",
            category = "Dam settings"
            )
        damSaveHeight.value = 0.5
        
        damWatersideSlope = arcpy.Parameter(
            displayName = "Slope steepness of water-side (1:X)",
            name = "damWatersideSlope",
            datatype = "GPDouble",
            parameterType = "Required",
            direction = "Input",
            category = "Dam settings"
            )
        damWatersideSlope.value = 3
        
        damWindwardSlope = arcpy.Parameter(
            displayName = "Slope steepness of windward-side (1:Y)",
            name = "damWindwardSlope",
            datatype = "GPDouble",
            parameterType = "Required",
            direction = "Input",
            category = "Dam settings"
            )
        damWindwardSlope.value = 2
        
        damFoundationDepth = arcpy.Parameter(
            displayName = "Dam foundation depth [m]",
            name = "damFoundationDepth",
            datatype = "GPDouble",
            parameterType = "Required",
            direction = "Input",
            category = "Dam settings"
            )
        damFoundationDepth.value = 1
        
        params = [inputProfiles, profileIDfieldName, inputDEM, targetWorkspace, saveProfiles, profilesOutputTable, depthStep, maxDamHeigth, useMinElevation, minElevationField,
                  damCrestWidth, damSaveHeight, damWatersideSlope, damWindwardSlope, damFoundationDepth, eraseIntermediateData]
        return params

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal validation is performed.  This method is called whenever a parameter has been changed.#"""
        if not parameters[4].hasBeenValidated:
            if parameters[4].value == 1:
                parameters[5].enabled = 1
                parameters[5].parameterType = "Required"
            if parameters[4].value == 0:
                parameters[5].enabled = 0
                parameters[5].parameterType = "Optional"
        if parameters[4].altered:
            if parameters[4].value:
                if parameters[4].value == 1:
                    parameters[5].enabled = 1
                    parameters[5].parameterType = "Required"
                if parameters[4].value == 0:
                    parameters[5].enabled = 0
                    parameters[5].parameterType = "Optional"      
        
        if not parameters[8].hasBeenValidated:
            if parameters[8].value == 1:
                parameters[9].enabled = 1
                parameters[9].parameterType = "Required"
            if parameters[8].value == 0:
                parameters[9].enabled = 0
                parameters[9].parameterType = "Optional"
                parameters[9].value = ""
        if parameters[8].altered:
            if parameters[8].value:
                if parameters[8].value == 1:
                    parameters[9].enabled = 1
                    parameters[9].parameterType = "Required"
                if parameters[8].value == 0:
                    parameters[9].enabled = 0
                    parameters[9].parameterType = "Optional"
                    parameters[9].value = ""
        
        if parameters[5].altered:
            ext = "csv"
            filename = os.path.basename(parameters[5].valueAsText)
            filext = filename.split(".")
            if len(filext) == 1:
                parameters[5].value = parameters[5].valueAsText+"."+ext
            if len(filext) > 1:
                extension = filext[-1]
                if extension != ext:
                    dirname = os.path.dirname(parameters[5].valueAsText)
                    filename = ".".join(filext[:-1])
                    parameters[5].value = os.path.join(dirname, filename)+"."+ext
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool parameter.  This method is called after internal validation.#"""

        return

    def execute(self, parameters, messages):
        """The source code of the tool."""
        sourceProfiles = parameters[0].valueAsText
        profileIDfieldName = parameters[1].valueAsText
        inputDEM = parameters[2].valueAsText
        targetWorkspace = parameters[3].valueAsText
        saveProfiles = parameters[4].value
        profilesOutputTable = parameters[5].valueAsText
        depthStep = parameters[6].value
        maxDamHeigth = parameters[7].value
        useStartElevation = parameters[8].value
        startElevationField = parameters[9].valueAsText
        damCrestWidth = locale.atof(str(parameters[10].valueAsText)) # type: float
        damSaveHeight = locale.atof(str(parameters[11].valueAsText)) # type: float
        damWatersideSlope = locale.atof(str(parameters[12].valueAsText)) # type: float
        damWindwardSlope = locale.atof(str(parameters[13].valueAsText)) # type: float
        damFoundationDepth = locale.atof(str(parameters[14].valueAsText)) # type: float
        eraseIntermediateData = parameters[15].value  # type: bool
        
        arcpy.env.workspace = targetWorkspace

        badProfiles = []

        cellSize = float(arcpy.management.GetRasterProperties(inputDEM, "CELLSIZEX").getOutput(0).replace(",", "."))
        #arcpy.AddMessage(cellSize)

        toBeDeleted = []

# CREATE AND FILL TASK SETTINGS TABLE ----------------------------------------------------------------------------------------
        taskSettingsFilePath = targetWorkspace+os.sep+taskSettingsFileName
        #if arcpy.Exists(taskSettingsFilePath):

        try:
            arcpy.management.CreateTable(targetWorkspace, taskSettingsFileName)
            arcpy.management.AddField(taskSettingsFilePath, IDfieldName, "TEXT", "", "", "", "", "", "", "")
            arcpy.management.AddField(taskSettingsFilePath, levelIntervalFieldName, "DOUBLE", "", "", "", "", "", "", "")
            arcpy.management.AddField(taskSettingsFilePath, maxDepthFieldName, "SHORT", "", "", "", "", "", "", "")
            arcpy.management.AddField(taskSettingsFilePath, startElevFieldFieldName, "TEXT", "", "", "", "", "", "", "")
            arcpy.management.AddField(taskSettingsFilePath, taskDEMFieldName, "TEXT", "", "", "", "", "", "", "")
            arcpy.management.AddField(taskSettingsFilePath, damCrestWidthFieldName, "DOUBLE", "", "", "", "", "", "", "")
            arcpy.management.AddField(taskSettingsFilePath, damSaveHeightFieldName, "DOUBLE", "", "", "", "", "", "", "")
            arcpy.management.AddField(taskSettingsFilePath, damWatersideSlopeFieldName, "DOUBLE", "", "", "", "", "", "", "")
            arcpy.management.AddField(taskSettingsFilePath, damWindwardSlopeFieldName, "DOUBLE", "", "", "", "", "", "", "")
            arcpy.management.AddField(taskSettingsFilePath, damFoundationDepthFieldName, "DOUBLE", "", "", "", "", "", "", "")

        except Exception as e:
            arcpy.AddMessage(e)
            arcpy.AddMessage("Failed creating task settings table '"+taskSettingsFilePath+"'")
        
        desc = arcpy.Describe(inputDEM)
        inputDEMfilePath = desc.catalogPath
        
        with arcpy.da.InsertCursor(taskSettingsFilePath, [IDfieldName, levelIntervalFieldName, maxDepthFieldName, startElevFieldFieldName, taskDEMFieldName, damCrestWidthFieldName, damSaveHeightFieldName, damWatersideSlopeFieldName, damWindwardSlopeFieldName, damFoundationDepthFieldName]) as table:
            table.insertRow([profileIDfieldName, depthStep, maxDamHeigth, startElevationField, inputDEMfilePath, damCrestWidth, damSaveHeight, damWatersideSlope, damWindwardSlope, damFoundationDepth])
            
# COPY THE PROFILES TO THE WORKING WORKSPACE --------------------------------------------------------------------------------------------
        inputProfiles = arcpy.management.CopyFeatures (sourceProfiles, targetWorkspace+os.sep+profilesFileName)
        
# GET THE STARTING ELEVATIONS IF USED
        if useStartElevation:
            startElevations = {}
            if startElevationField in (None, "", " ", "#"):
                arcpy.AddError("The minimum elevation field must be selected")
                pass
            else:
                with arcpy.da.SearchCursor(targetWorkspace+os.sep+profilesFileName, [profileIDfieldName, startElevationField]) as profiles:
                    for pr in profiles:
                        startElevations.update({pr[0]:pr[1]})
        
# GETING THE PROFILES ELEVATION ----------------------------------------------------------------------------------------------
        profileElevFilePath = os.path.join(targetWorkspace, profileElevFileName)
        
        try:
            startTime = datetime.datetime.now()
            arcpy.AddMessage("Getting elevations of profiles ...") 
            arcpy.ddd.InterpolateShape(inputDEM, inputProfiles, profileElevFilePath, "", "", "LINEAR", "DENSIFY")
        except Exception as e:
            arcpy.AddError(e.message)
            arcpy.AddError("Extracting elevations of profiles failed!")  
        else:
            arcpy.AddMessage("    ... successfull. ("+str(datetime.datetime.now()-startTime)+")") 
            toBeDeleted.append(profileElevFilePath)

        #profilePoints = {profile ID: [point, ...]}
        profilePoints = {}
        #profileTable {profile ID: [[point position cumulative, point elevation, distance from previous point], ... ]}
        profileTable = {}
        
        #shapeFieldName = getShapeFieldName(profileElevFilePath)
        shapeFieldName = "SHAPE@"
        
        arcpy.AddMessage("Getting cross-sections ...")
        startTime = datetime.datetime.now()
        with arcpy.da.SearchCursor(profileElevFilePath, [profileIDfieldName, shapeFieldName]) as profiles:
            for row in profiles:
                # points to be used in further processing
                pointsForProfile = []
                # point to be written into the output table
                pointsForTable = []
                # Print the current multipoint's ID
                #arcpy.AddMessage("Feature {}:".format(row[0]))
                partnum = 0

                # Step through each part of the feature
                for part in row[1]:
                    pid = row[0]
                    numPoints = len(part)
                    p = 0
                    cumPos = 0
                    # Step through each vertex in the feature
                    for pnt in part:
                        if pnt:
                            pointsForProfile.append(pnt)
                            if p == 0:
                                pointsForTable.append([0, pnt.Z, 0])
                                prevPnt = pnt
                            else:
                                dist = pow(pow(pnt.X - prevPnt.X, 2) + pow(pnt.Y - prevPnt.Y, 2), 0.5)
                                cumPos += dist
                                pointsForTable.append([cumPos, pnt.Z, dist])
                                prevPnt = pnt
                        else:
                            pass
                        p += 1
                        
                    partnum += 1

                profilePoints.update({pid:pointsForProfile})
                profileTable.update({pid:pointsForTable})
        arcpy.AddMessage("    ... successfull. ("+str(datetime.datetime.now()-startTime)+")") 
        #arcpy.AddMessage(str(profileTable.keys()))
        #arcpy.AddMessage(str(profileTable))

        # write the profiles' cross-sections into the file
        if saveProfiles:
            arcpy.AddMessage("Saving cross-sections ...")
            startTime = datetime.datetime.now()

            file = open(profilesOutputTable, 'w')
            for profileID in profileTable.keys():
                profile = profileTable[profileID]
                file.write("Profile ID:;"+str(profileID)+"\n")
                #file.write("\n")
                for p in profile:
                    file.write(str(p[0])+";")
                file.write("\n")
                for p in profile:
                    file.write(str(p[1])+";")
                file.write("\n")
                for p in profile:
                    file.write(str(p[2])+";")
                file.write("\n\n\n")
            file.close()
            arcpy.AddMessage("    ... successfull. ("+str(datetime.datetime.now()-startTime)+")")


# ANALYSING THE PROFILES' CROSS-SECTIONS -------------------------------------------------------------------------------------

        # create the table for general properties of all profiles
        profilesPropsTablePath = targetWorkspace+os.sep+profilesPropsTableName

        try:
            arcpy.management.CreateTable(targetWorkspace, profilesPropsTableName)
            arcpy.management.AddField(profilesPropsTablePath, profileIDfieldName, "LONG", "", "", "", "", "", "", "")
            arcpy.management.AddField(profilesPropsTablePath, minElevFieldName, "DOUBLE", "", "", "", "", "", "", "")
            arcpy.management.AddField(profilesPropsTablePath, maxLeftFieldName, "DOUBLE", "", "", "", "", "", "", "")
            arcpy.management.AddField(profilesPropsTablePath, maxRightFieldName, "DOUBLE", "", "", "", "", "", "", "")
            arcpy.management.AddField(profilesPropsTablePath, rangeFieldName, "DOUBLE", "", "", "", "", "", "", "")
            arcpy.management.AddField(profilesPropsTablePath, maxLevelFieldName, "SHORT", "", "", "", "", "", "", "")
            arcpy.management.AddField(profilesPropsTablePath, minLevelFieldName, "SHORT", "", "", "", "", "", "", "")
            if useStartElevation:
                arcpy.management.AddField(profilesPropsTablePath, startElevFieldName, "DOUBLE", "", "", "", "", "", "", "")
                
        except Exception as e:
            arcpy.AddMessage(e)
            arcpy.AddMessage("Failed creating profile properties table '"+profilesPropsTablePath+"'")

        # calculate the profiles properties and write it to the table
        profileMins = {}
        profileLimits = {}
        profileMinPoints = {}
        profileExtremesIndices = {}
        with arcpy.da.InsertCursor(profilesPropsTablePath, [profileIDfieldName, minElevFieldName, maxLeftFieldName, maxRightFieldName, rangeFieldName]) as table:
            try:
                arcpy.AddMessage("Getting cross-sections properties ...")
                startTime = datetime.datetime.now()
                for profileID in profilePoints.keys():
                    #arcpy.AddMessage("  Profile #"+str(profileID))
                    profile = profilePoints[profileID]
                    maxLeft = 0
                    maxRight = 0
                    minimum = 10000
                    numPoints = len(profile)
                    p = 0
                    for point in profile:
                        #arcpy.AddMessage(str(point))
                        if p == 0:
                            prevPoint = point
                        else:
                            if (point.Z < prevPoint.Z) and (point.Z < minimum):
                                minimum = point.Z
                                minIndex = p
                            prevPoint = point
                        p += 1
                    #arcpy.AddMessage("    min index: "+str(minIndex))

                    p = 0
                    for point in profile:
                        if p == 0:
                            prevPoint = point
                            maxLeft = point.Z
                            LmaxIndex = 0
                        else:
                            if p < minIndex:
                                #arcpy.AddMessage("    "+str(p)+" - "+str(point.Z))
                                if (point.Z > prevPoint.Z) and (point.Z > maxLeft):
                                    #arcpy.AddMessage("    left maximum: "+str(p)+" - "+str(point.Z))
                                    maxLeft = point.Z
                                    LmaxIndex = p
                            if p > minIndex:
                                #arcpy.AddMessage("    (point.Z "+str(point.Z)+")")
                                if (point.Z > prevPoint.Z) and (point.Z > maxRight):
                                    #arcpy.AddMessage("    right maximum: "+str(p)+" - "+str(point.Z))
                                    maxRight = point.Z
                                    RmaxIndex = p
                            prevPoint = point
                        p += 1
                    
                    maxAbs = min(maxLeft, maxRight)
                    elevRange = maxAbs - minimum
                    table.insertRow([profileID, minimum, maxLeft, maxRight, elevRange])

                    # store the values in dictionaries for later use
                    profileMins.update({profileID: [minIndex, minimum]})
                    profileLimits.update({profileID: [minimum, maxAbs]})
                    profileMinPoints.update({profileID: profile[minIndex]})
                    profileExtremesIndices.update({profileID:[LmaxIndex, minIndex, RmaxIndex]})
                    
            except Exception as e:
                arcpy.AddMessage(e)
                arcpy.AddMessage("Failed getting cross-sections properties.")
            else:
                arcpy.AddMessage("    ... successfull. ("+str(datetime.datetime.now()-startTime)+")")

# WRITE THE STARTING ELEVATIONS INTO THE PROPERTIES TABLE --------------------------------------------------------------------------------------------------
        if useStartElevation:
            with arcpy.da.UpdateCursor (profilesPropsTablePath, [profileIDfieldName, startElevFieldName]) as table:
                for row in table:
                    row[1] = startElevations[row[0]]
                    table.updateRow(row)

# CREATING MINIMUM ELEVATION POINTS FEATURE CLASS FOR PROFILES ----------------------------------------------------------------------------------------------
        arcpy.AddMessage("Getting lowest points on profiles ...")
        startTime = datetime.datetime.now()
        
        profileMinPointsFilePath = targetWorkspace+os.sep+profileMinPointsFileName
        
        try:
            arcpy.management.CreateFeatureclass(targetWorkspace, profileMinPointsFileName, "POINT", "", "DISABLED", "ENABLED", sr)
            arcpy.management.AddField(profileMinPointsFilePath, profileIDfieldName, "LONG", "", "", "", "", "", "", "")
        except Exception as e:
            arcpy.AddMessage(e)
            arcpy.AddMessage("Failed creating min points feature class.'"+profileMinPointsFilePath+"'")
        
        #shapeFieldName = getShapeFieldName(profileMinPointsFilePath)
        pointsCursor = arcpy.da.InsertCursor(profileMinPointsFilePath, [shapeFieldName, profileIDfieldName])
        
        try:
            #arcpy.AddMessage("  Writing min points feature class.")
            for key in profileMinPoints.keys():
                point = profileMinPoints.get(key)
                #arcpy.AddMessage(str(point))
                pointsCursor.insertRow([point, key])
        except Exception as e:
            arcpy.AddMessage(e)
            arcpy.AddMessage("  Failed writing min points feature class.")
        else:
            arcpy.AddMessage("    ... successfull. ("+str(datetime.datetime.now()-startTime)+")")
        finally:
            del pointsCursor         
                       
# SPLITTING PROFILES BY MINIMUM POINTS ---------------------------------------------------------------------------------------------------------------------
        profilesSplitPoints = {}
        for pid in profileExtremesIndices.keys():
            LmaxIndex = profileExtremesIndices[pid][0]
            minIndex = profileExtremesIndices[pid][1]
            RmaxIndex = profileExtremesIndices[pid][2]
            
            
            pPoints = profilePoints[pid]
            
            leftPartPoints = pPoints[LmaxIndex:(minIndex+1)]
            rightPartPoints = pPoints[minIndex:(RmaxIndex+1)]
            profilesSplitPoints.update({pid:[leftPartPoints, rightPartPoints]})

        
        arcpy.AddMessage("Splitting profiles ...")
        startTime = datetime.datetime.now()
        profileSplitFileName = "profiles_split"
        profileSplitFilePath = targetWorkspace+os.sep+profileSplitFileName
        
        try:
            arcpy.management.CreateFeatureclass(targetWorkspace, profileSplitFileName, "POLYLINE", "", "DISABLED", "ENABLED", sr)
            arcpy.management.AddField(profileSplitFilePath, profileIDfieldName, "LONG", "", "", "", "", "", "", "")
            arcpy.management.AddField(profileSplitFilePath, "side", "TEXT", "", "", "", "", "", "", "")
            with arcpy.da.InsertCursor(profileSplitFilePath, [shapeFieldName, profileIDfieldName, sideFieldName]) as profilesSplitCursor:
                for pid in profilesSplitPoints.keys():
                    profilesSplitCursor.insertRow([arcpy.Polyline(arcpy.Array(profilesSplitPoints[pid][0]), sr), pid, "L"])
                    profilesSplitCursor.insertRow([arcpy.Polyline(arcpy.Array(profilesSplitPoints[pid][1]), sr), pid, "R"])
        except Exception as e:
            arcpy.AddMessage(e)
            arcpy.AddMessage("Failed creating split profiles feature class.'"+profileSplitFilePath+"'")
        else:
            arcpy.AddMessage("    ... successfull. ("+str(datetime.datetime.now()-startTime)+")")
        
# CREATING THE INSIDE POINTS --------------------------------------------------------------------------------------------------------------------
        # the "inside point" is a point derived from the dam profile
        # it is the minimum elevation point on the profile offset by a distance based on the dam body geometry in the uphill direction

        arcpy.AddMessage("Creating inside points ...")
        startTime = datetime.datetime.now()
        
        insidePointsFilePath = targetWorkspace+os.sep+insidePointsFileName
        
        try:
            arcpy.management.CreateFeatureclass (targetWorkspace, insidePointsFileName, "POINT", "", "DISABLED", "ENABLED", sr)
            arcpy.management.AddField(insidePointsFilePath, profileIDfieldName, "LONG", "", "", "", "", "", "", "")
        except Exception as e:
            arcpy.AddMessage(e)
            arcpy.AddMessage("Failed creating inside points feature class.'"+insidePointsFilePath+"'")
            
        insidePoints = {}
        # the distence from the profile to the inside point is dependent on the dam body geometry
        dist = 2*(damCrestWidth/2 + damSaveHeight * damWatersideSlope)
        
        for pid in profilesSplitPoints.keys():
            i = 0
            # calculate the inside points localtion from the first two points of the second half of the split profile
            for pnt in profilesSplitPoints[pid][1]:
                if i == 0:
                    prevPnt = pnt
                elif i == 1:
                    dy = pnt.Y - prevPnt.Y  
                    dx = pnt.X - prevPnt.X
                    # calculate the angle of the line between the first two points
                    segmentAngle = math.degrees(math.atan2(dy,dx))
                    # calculate the points coordinates from the distance and the perpendicular line
                    newPnt = arcpy.Point(prevPnt.X - (math.cos(math.radians(segmentAngle+90))*dist), prevPnt.Y - (math.sin(math.radians(segmentAngle+90))*dist))
                    insidePoints.update({pid:newPnt})
                i += 1
        # and write the inside points in feature class
        try:
            with arcpy.da.InsertCursor(insidePointsFilePath, [profileIDfieldName, shapeFieldName]) as pointsCursor:
                for key in insidePoints.keys():
                    pointsCursor.insertRow([key, insidePoints[key]])
        except Exception as e:
            arcpy.AddMessage(e)
            arcpy.AddMessage("  Failed writing inside points feature class.")
        else:
            arcpy.AddMessage("    ... successfull. ("+str(datetime.datetime.now()-startTime)+")")
         
# CREATING CONTOUR LINES LISTS FOR PROFILES -----------------------------------------------------------------------------------------------------------------------
        arcpy.AddMessage("Calculating contour line heights for profiles ...")
        startTime = datetime.datetime.now()
        profileContourElev = {}
        for profileID in profileLimits.keys():
            contourList = []
            if profileID not in badProfiles:
                #arcpy.AddMessage("  profile #"+str(profileID))
                if useStartElevation:
                    if startElevations[profileID] != 0:
                        minElev = startElevations[profileID]
                    else:
                        minElev = profileLimits[profileID][0]
                else:
                    minElev = profileLimits[profileID][0]

                maxElev = profileLimits[profileID][1]
                damHeight = depthStep
                while damHeight <= maxDamHeigth:
                    elevation = minElev + damHeight - damSaveHeight
                    if elevation > minElev and elevation <= maxElev:
                        contourList.append(elevation)                
                    damHeight += depthStep
                if len(contourList) == 0:
                    arcpy.AddWarning("Profile #"+str(profileID)+" doesn't have sufficient elevation difference!")
                    badProfiles.append(profileID)
            profileContourElev.update({profileID: contourList})

        arcpy.AddMessage("    ... successfull. ("+str(datetime.datetime.now()-startTime)+")")
        #arcpy.AddMessage(str(profileContourElev))

# CREATING THE DAM BODIES -------------------------------------------------------------------------------------------------------------------------------------------
        arcpy.AddMessage("Creating dam bodies ...")
        startTime = datetime.datetime.now()
        # get the points of the original profile for dam construction
        profilePoints = {}
        with arcpy.da.SearchCursor(inputProfiles, [profileIDfieldName, shapeFieldName]) as profiles:
            for row in profiles:
                points = []
                # Step through each part of the feature
                for part in row[1]:
                    pid = row[0]
                    # Step through each vertex in the feature
                    for pnt in part:
                        if pnt:
                            points.append(pnt)

                        else:
                            # If pnt is None, this represents an interior ring
                            pass
                        
                profilePoints.update({pid:points})
                    
        newLines = []
        # characteristic lines of the dam body (edges of the dams crest)
        charLines = {}
        # waterline (where the flooded area meets the dam body)
        waterLines = {}
        # line between waterline and dam crest used to remove unwanted polygons from selection
        eraseLines = {}
        
        distCrest = damCrestWidth/2
        # how high will the dam 3D model be? [m]
        damModelHeight = 2 * maxDamHeigth
        elevAbove = damSaveHeight
        elevBelow = damSaveHeight - damModelHeight
        # distance upstream - water side - heel
        distUp = damModelHeight * damWatersideSlope + distCrest
        # distance downstream - windward side - toe
        distDown = damModelHeight * damWindwardSlope + distCrest
        # distance from the mid-line to waterline (where the flooded area meets the dam body)
        distWLine = distCrest + damSaveHeight * damWatersideSlope
        distELine = distWLine - 0.2
        # added length to the water lines - to make sure that the contour and the water line intersect
        wLineOverLength = 50

        numDigits = max(len(str(max(profilePoints.keys()))), 2)
        
        for pid in profilePoints.keys():
            # all of dam body significant points
            points3D = []
            # points of the upstream crest line
            uCrestPoints = []
            # points of the downstream crest line
            dCrestPoints = []
            # points of the upstream heel line
            upPoints = []
            # points of the downstream heel line
            downPoints = []
            # points of the water line - where the flooded area meets the dam body
            wLinePoints = []
            eLinePoints = []
            
            for i in range(len(profilePoints[pid])):
                pnts = profilePoints[pid]
                # for the very first point
                if i == 0:
                    #arcpy.AddMessage("processing first point")
                    dy = pnts[i+1].Y - pnts[i].Y  
                    dx = pnts[i+1].X - pnts[i].X
                    segmentAngle = math.degrees(math.atan2(dy,dx))
                    segmentLength = pow(pow((pnts[i+1].X - pnts[i].X), 2) + pow(pnts[i+1].Y - pnts[i].Y, 2), 0.5)
                    # segment's unit vector
                    dyu = dy/segmentLength
                    dxu = dx/segmentLength

                    uCrest = arcpy.Point(pnts[i].X + (math.cos(math.radians(segmentAngle-90))*distCrest), pnts[i].Y + (math.sin(math.radians(segmentAngle-90))*distCrest), elevAbove)
                    dCrest = arcpy.Point(pnts[i].X - (math.cos(math.radians(segmentAngle-90))*distCrest), pnts[i].Y - (math.sin(math.radians(segmentAngle-90))*distCrest), elevAbove)
                    left = arcpy.Point(pnts[i].X + (math.cos(math.radians(segmentAngle-90))*distUp), pnts[i].Y + (math.sin(math.radians(segmentAngle-90))*distUp), elevBelow)
                    right = arcpy.Point(pnts[i].X - (math.cos(math.radians(segmentAngle-90))*distDown), pnts[i].Y - (math.sin(math.radians(segmentAngle-90))*distDown), elevBelow)                   
                    #add one point to the water line before the beginning
                    wFirstPoint = arcpy.Point(pnts[i].X + (math.cos(math.radians(segmentAngle-90))*distWLine), pnts[i].Y + (math.sin(math.radians(segmentAngle-90))*distWLine), 0)
                    wLine = [arcpy.Point(wFirstPoint.X-dxu*wLineOverLength, wFirstPoint.Y-dyu*wLineOverLength, 0)]
                    wLine.append(wFirstPoint)
                    
                    eLine = arcpy.Point(pnts[i].X + (math.cos(math.radians(segmentAngle-90))*distELine), pnts[i].Y + (math.sin(math.radians(segmentAngle-90))*distELine), 0)

                # for all the other points
                elif i < len(profilePoints[pid])-1:
                    dy1 = pnts[i].Y - pnts[i-1].Y 
                    dx1 = pnts[i].X - pnts[i-1].X
                    dy2 = pnts[i+1].Y - pnts[i].Y  
                    dx2 = pnts[i+1].X - pnts[i].X
                    
                    segmentAngle1 = math.degrees(math.atan2(dy1,dx1))
                    segmentAngle2 = math.degrees(math.atan2(dy2,dx2))
                    
                    uCrest1 = arcpy.Point(pnts[i-1].X + (math.cos(math.radians(segmentAngle1-90))*distCrest), pnts[i-1].Y + (math.sin(math.radians(segmentAngle1-90))*distCrest))
                    dCrest1 = arcpy.Point(pnts[i-1].X - (math.cos(math.radians(segmentAngle1-90))*distCrest), pnts[i-1].Y - (math.sin(math.radians(segmentAngle1-90))*distCrest))
                    left1 = arcpy.Point(pnts[i-1].X + (math.cos(math.radians(segmentAngle1-90))*distUp), pnts[i-1].Y + (math.sin(math.radians(segmentAngle1-90))*distUp))
                    right1 = arcpy.Point(pnts[i-1].X - (math.cos(math.radians(segmentAngle1-90))*distDown), pnts[i-1].Y - (math.sin(math.radians(segmentAngle1-90))*distDown))
                    wLine1 = arcpy.Point(pnts[i-1].X + (math.cos(math.radians(segmentAngle1-90))*distWLine), pnts[i-1].Y + (math.sin(math.radians(segmentAngle1-90))*distWLine))
                    eLine1 = arcpy.Point(pnts[i-1].X + (math.cos(math.radians(segmentAngle1-90))*distELine), pnts[i-1].Y + (math.sin(math.radians(segmentAngle1-90))*distELine))
                    
                    uCrest2 = arcpy.Point(pnts[i].X + (math.cos(math.radians(segmentAngle2-90))*distCrest), pnts[i].Y + (math.sin(math.radians(segmentAngle2-90))*distCrest))
                    dCrest2 = arcpy.Point(pnts[i].X - (math.cos(math.radians(segmentAngle2-90))*distCrest), pnts[i].Y - (math.sin(math.radians(segmentAngle2-90))*distCrest))
                    left2 = arcpy.Point(pnts[i].X + (math.cos(math.radians(segmentAngle2-90))*distUp), pnts[i].Y + (math.sin(math.radians(segmentAngle2-90))*distUp))
                    right2 = arcpy.Point(pnts[i].X - (math.cos(math.radians(segmentAngle2-90))*distDown), pnts[i].Y - (math.sin(math.radians(segmentAngle2-90))*distDown))
                    wLine2 = arcpy.Point(pnts[i].X + (math.cos(math.radians(segmentAngle2-90))*distWLine), pnts[i].Y + (math.sin(math.radians(segmentAngle2-90))*distWLine))
                    eLine2 = arcpy.Point(pnts[i].X + (math.cos(math.radians(segmentAngle2-90))*distELine), pnts[i].Y + (math.sin(math.radians(segmentAngle2-90))*distELine))
                    
                    uCrestY = (uCrest1.Y*(dx1/dy1)-uCrest2.Y*(dx2/dy2)+uCrest2.X-uCrest1.X)*math.pow(((dx1/dy1)-(dx2/dy2)), -1)
                    uCrestX = uCrest2.X+(dx2/dy2)*(uCrestY-uCrest2.Y)
                    uCrest = arcpy.Point(uCrestX, uCrestY, elevAbove)
                    
                    dCrestY = (dCrest1.Y*(dx1/dy1)-dCrest2.Y*(dx2/dy2)+dCrest2.X-dCrest1.X)*math.pow(((dx1/dy1)-(dx2/dy2)), -1)
                    dCrestX = dCrest2.X+(dx2/dy2)*(dCrestY-dCrest2.Y)
                    dCrest = arcpy.Point(dCrestX, dCrestY, elevAbove)
                    
                    leftY = (left1.Y*(dx1/dy1)-left2.Y*(dx2/dy2)+left2.X-left1.X)*math.pow(((dx1/dy1)-(dx2/dy2)), -1)
                    leftX = left2.X+(dx2/dy2)*(leftY-left2.Y)
                    left = arcpy.Point(leftX, leftY, elevBelow)
                    
                    rightY = (right1.Y*(dx1/dy1)-right2.Y*(dx2/dy2)+right2.X-right1.X)*math.pow(((dx1/dy1)-(dx2/dy2)), -1)
                    rightX = right2.X+(dx2/dy2)*(rightY-right2.Y)
                    right = arcpy.Point(rightX, rightY, elevBelow)
                    
                    wLineY = (wLine1.Y*(dx1/dy1)-wLine2.Y*(dx2/dy2)+wLine2.X-wLine1.X)*math.pow(((dx1/dy1)-(dx2/dy2)), -1)
                    wLineX = wLine2.X+(dx2/dy2)*(wLineY-wLine2.Y)
                    wLine = [arcpy.Point(wLineX, wLineY, 0)]
                    
                    eLineY = (eLine1.Y*(dx1/dy1)-eLine2.Y*(dx2/dy2)+eLine2.X-eLine1.X)*math.pow(((dx1/dy1)-(dx2/dy2)), -1)
                    eLineX = eLine2.X+(dx2/dy2)*(eLineY-eLine2.Y)
                    eLine = arcpy.Point(eLineX, eLineY, 0)
                    
                # for the very last point
                else:
                    #arcpy.AddMessage("processing last point")
                    dy = pnts[i].Y - pnts[i-1].Y  
                    dx = pnts[i].X - pnts[i-1].X
                    segmentAngle = math.degrees(math.atan2(dy,dx))  
                    segmentLength = pow(pow((pnts[i].X - pnts[i-1].X), 2) + pow(pnts[i].Y - pnts[i-1].Y, 2), 0.5)
                    # segment's unit vector
                    dyu = dy/segmentLength
                    dxu = dx/segmentLength
                    
                    uCrest = arcpy.Point(pnts[i].X + (math.cos(math.radians(segmentAngle-90))*distCrest), pnts[i].Y + (math.sin(math.radians(segmentAngle-90))*distCrest), elevAbove)
                    dCrest = arcpy.Point(pnts[i].X - (math.cos(math.radians(segmentAngle-90))*distCrest), pnts[i].Y - (math.sin(math.radians(segmentAngle-90))*distCrest), elevAbove)
                    left = arcpy.Point(pnts[i].X + (math.cos(math.radians(segmentAngle-90))*distUp), pnts[i].Y + (math.sin(math.radians(segmentAngle-90))*distUp), elevBelow)
                    right = arcpy.Point(pnts[i].X - (math.cos(math.radians(segmentAngle-90))*distDown), pnts[i].Y - (math.sin(math.radians(segmentAngle-90))*distDown), elevBelow)
                    
                    # add one more point after the end
                    wLastPoint = arcpy.Point(pnts[i].X + (math.cos(math.radians(segmentAngle-90))*distWLine), pnts[i].Y + (math.sin(math.radians(segmentAngle-90))*distWLine), 0)
                    wLine = [wLastPoint]
                    wLine.append(arcpy.Point(wLastPoint.X+dxu*wLineOverLength, wLastPoint.Y+dyu*wLineOverLength, 0))
                    
                    eLine = arcpy.Point(pnts[i].X + (math.cos(math.radians(segmentAngle-90))*distELine), pnts[i].Y + (math.sin(math.radians(segmentAngle-90))*distELine), 0)
                    
                uCrestPoints.append(uCrest)
                dCrestPoints.append(dCrest)
                upPoints.append(left)
                downPoints.append(right)
                points3D.extend([uCrest, dCrest, left, right])
                wLinePoints.extend(wLine)
                eLinePoints.append(eLine)
                
                postfix = str(pid).zfill(numDigits)
            
            # create the dam body points feature class
            
            arcpy.management.CopyFeatures(arcpy.Multipoint(arcpy.Array(points3D), sr, True, False), targetWorkspace+os.sep+damPointsNamePrefix+postfix)
            toBeDeleted.append(targetWorkspace+os.sep+damPointsNamePrefix+postfix)

            newLines.append(arcpy.Polyline(arcpy.Array(uCrestPoints),sr))
            newLines.append(arcpy.Polyline(arcpy.Array(dCrestPoints),sr))
            newLines.append(arcpy.Polyline(arcpy.Array(upPoints),sr))
            newLines.append(arcpy.Polyline(arcpy.Array(downPoints),sr))
            charLines.update({pid:[arcpy.Polyline(arcpy.Array(uCrestPoints)), arcpy.Polyline(arcpy.Array(dCrestPoints)), arcpy.Polyline(arcpy.Array(upPoints)), arcpy.Polyline(arcpy.Array(downPoints))]})
            waterLines.update({pid: arcpy.Polyline(arcpy.Array(wLinePoints))})
            eraseLines.update({pid: arcpy.Polyline(arcpy.Array(eLinePoints))})
            
        # write the dam body lines feature class
        charLinesFilePath = targetWorkspace+os.sep+charLinesFileName
        try:
            arcpy.management.CreateFeatureclass (targetWorkspace, charLinesFileName, "POLYLINE", "", "DISABLED", "ENABLED", sr)
            arcpy.management.AddField(charLinesFilePath, profileIDfieldName, "LONG", "", "", "", "", "", "", "")
        except Exception as e:
            arcpy.AddMessage(e)
            arcpy.AddMessage("Failed creating characteristical lines feature class.'"+charLinesFileName+"'")
        
        with arcpy.da.InsertCursor(charLinesFilePath, [profileIDfieldName, shapeFieldName]) as table:
            for pid in charLines.keys():
                for line in charLines[pid]:
                    table.insertRow([pid, line])
        
        # write the lines where the water meets the dam body
        waterLinesFilePath = targetWorkspace+os.sep+waterLinesFileName
        try:
            arcpy.management.CreateFeatureclass (targetWorkspace, waterLinesFileName, "POLYLINE", "", "DISABLED", "ENABLED", sr)
            arcpy.management.AddField(waterLinesFilePath, profileIDfieldName, "LONG", "", "", "", "", "", "", "")
        except Exception as e:
            arcpy.AddMessage(e)
            arcpy.AddMessage("Failed creating water lines feature class.'"+waterLinesFilePath+"'")
        
        with arcpy.da.InsertCursor(waterLinesFilePath, [profileIDfieldName, shapeFieldName]) as table:
            for pid in waterLines.keys():
                table.insertRow([pid, waterLines[pid]])
                
        # write the erase lines
        eraseLinesFilePath = targetWorkspace+os.sep+"erase_lines"
        try:
            arcpy.management.CreateFeatureclass (targetWorkspace, "erase_lines", "POLYLINE", "", "DISABLED", "ENABLED", sr)
            arcpy.management.AddField(eraseLinesFilePath, profileIDfieldName, "LONG", "", "", "", "", "", "", "")
        except Exception as e:
            arcpy.AddMessage(e)
            arcpy.AddMessage("Failed creating water lines feature class.'"+eraseLinesFilePath+"'")
        
        with arcpy.da.InsertCursor(eraseLinesFilePath, [profileIDfieldName, shapeFieldName]) as table:
            for pid in waterLines.keys():
                table.insertRow([pid, waterLines[pid]])
        toBeDeleted.append(eraseLinesFilePath)
        
        arcpy.AddMessage("    ... successfull. ("+str(datetime.datetime.now()-startTime)+")")


# CREATE THE MAXIMUM WATER LEVEL FLOODED AREA FOR EACH PROFILE FOR DEM CLIPPING ---------------------------------------------------------------
        arcpy.AddMessage("Creating DEM for each profile ...")
        startTime = datetime.datetime.now()

        numDigits = max(len(str(max(profileLimits.keys()))), 2)

        arcpy.management.MakeFeatureLayer (inputProfiles, "profilesLayer")
        arcpy.management.MakeFeatureLayer (insidePointsFilePath, "insidePointsLayer")
        
        
        for profileID in profileLimits.keys():
            if profileID not in badProfiles:
                #arcpy.AddMessage("  getting maximum flooded area for proile #"+str(profileID)+"...")
                #arcpy.AddMessage("  max water level elevation: "+str(profileLimits[profileID][1]))
                postfix = str(profileID).zfill(numDigits)   
                maxContourFilePath = targetWorkspace+os.sep+prefixMaxCont+postfix
                arcpy.sa.ContourList (inputDEM, maxContourFilePath, profileLimits[profileID][1]-0.5)
                toBeDeleted.append(maxContourFilePath)
                
                where = arcpy.AddFieldDelimiters("profilesLayer", profileIDfieldName)+" = "+str(profileID)
                arcpy.management.SelectLayerByAttribute ("profilesLayer", "NEW_SELECTION", where)
                where2 = arcpy.AddFieldDelimiters("insidePointsLayer", profileIDfieldName)+" = "+str(profileID)
                arcpy.management.SelectLayerByAttribute ("insidePointsLayer", "NEW_SELECTION", where2)
                
                contPolyFileName = prefixMaxContPoly+postfix
                contPolyFilePath = targetWorkspace+os.sep+contPolyFileName
                arcpy.management.FeatureToPolygon([maxContourFilePath, "profilesLayer"], contPolyFilePath, "", "ATTRIBUTES")
                toBeDeleted.append(contPolyFilePath)
                
                with arcpy.da.SearchCursor(contPolyFilePath, [shapeFieldName]) as maxPolys:
                    c = 0
                    for poly in maxPolys:
                        if insidePoints[profileID].within(poly[0]):
                            #arcpy.AddMessage("True")
                            buffPoly = poly[0].buffer(50)
                            extent = buffPoly.extent
                            #arcpy.AddMessage(str(extent))
                            c += 1
                        else:
                            #arcpy.AddMessage("false")
                            pass
                    if c == 0:
                        arcpy.AddWarning("  profile #"+str(profileID)+" - no polygon created from maximum level contour and dam profile.")
                        arcpy.AddWarning("  Check dam profile orientation, length, elevation profile and DEM extent.")
                        badProfiles.append(profileID)
                    elif c > 1:
                        arcpy.AddWarning("  profile #"+str(profileID)+" - more than one polygon created from maximum level contour and dam profile!")
                        arcpy.AddWarning("  Check dam profile geometry.")
                        badProfiles.append(profileID)
                    else:
                        DEMfileName = DEMfileNamePrefix+postfix
                        DEMfilePath = targetWorkspace+os.sep+DEMfileName
                        arcpy.management.Clip (inputDEM, str(extent.XMin)+" "+str(extent.YMin)+" "+str(extent.XMax)+" "+str(extent.YMax), DEMfilePath)
                        arcpy.AddMessage("  profile #"+str(profileID)+" done")
            else:
                arcpy.AddMessage("  profile #"+str(profileID)+" not processed")
        arcpy.AddMessage("    ... successfull. ("+str(datetime.datetime.now()-startTime)+")")

# CREATING DEMs FOR THE DAM BODIES --------------------------------------------------------------------------------------------------------------------------
        arcpy.AddMessage("Creating dam bodies models ...")
        startTime = datetime.datetime.now()
        numDigits = max(len(str(max(profileContourElev.keys()))), 2)
        
        # create the dir for the TINs
        tindir = os.path.dirname(targetWorkspace)+os.sep+"tins"
        if not os.path.exists(tindir):
            os.makedirs(tindir)

        # arcpy.AddMessage(str(os.access(tindir, os.W_OK)))
        # checkt if the TIN target directory has write permission
        if os.access(tindir, os.W_OK):
            # convert dam body points to TIN and raster
            for profileID in profileContourElev.keys():
                arcpy.AddMessage("  profile #"+str(profileID))
                postfix = str(profileID).zfill(numDigits)
                arcpy.env.snapRaster =  targetWorkspace+os.sep+DEMfileNamePrefix+postfix
                #arcpy.AddMessage(tindir+os.sep+"damtin_"+postfix)

                arcpy.ddd.CreateTin(tindir+os.sep+"damtin_"+postfix, sr, targetWorkspace+os.sep+damPointsNamePrefix+postfix+" Shape.Z Mass_Points <None>", "CONSTRAINED_DELAUNAY")

                toBeDeleted.append(tindir+os.sep+"damtin_"+postfix)
                arcpy.AddMessage("    TIN surface model done")
                arcpy.ddd.TinRaster(tindir+os.sep+"damtin_"+postfix, targetWorkspace+os.sep+"damRaster_"+postfix, "FLOAT", "LINEAR", "CELLSIZE "+str(cellSize))
                toBeDeleted.append(targetWorkspace+os.sep+"damRaster_"+postfix)
                arcpy.AddMessage("    conversion to raster done")
            arcpy.AddMessage("    ... successfull. ("+str(datetime.datetime.now()-startTime)+")")
        else:
            arcpy.AddError("The TIN target directory '"+tindir+"' is forbidden for writing. Can't proceed with dam body processing.")
            return

# CREATING CONTOUR LINES FOR INDIVIDUAL DAM PROFILES -------------------------------------------------------------------------------------------------------
        arcpy.AddMessage("Creating flooded areas for profiles ...")
        startTime = datetime.datetime.now()

        numDigits = max(len(str(max(profileContourElev.keys()))), 2)

        arcpy.management.MakeFeatureLayer(inputProfiles, "profilesLayer")
        arcpy.management.MakeFeatureLayer(charLinesFilePath, "charLinesLayer")
        arcpy.management.MakeFeatureLayer(waterLinesFilePath, "waterLinesLayer")
        arcpy.management.MakeFeatureLayer(eraseLinesFilePath, "eraseLinesLayer")

        toBeDeleted.extend(["profilesLayer", "insidePointsLayer", "charLinesLayer", "waterLinesLayer", "eraseLinesLayer"])

        # minimal area of small ponds
        areaThreshold = cellSize*cellSize
        # list of minimum wate levels
        minLevels = {}
        # list of maximum water levels of profiles to be stored into profile properties table
        maxLevels = {}
        
        maxLevel = 0
        for profileID in profileContourElev.keys():
            if profileID not in badProfiles:
                arcpy.AddMessage("\nProcessing profile #"+str(profileID)+":")

                # properties of flooded areas of a single profile
                props = []
                # first successfully created water level
                minLevel = 1
                # maximum successfully created water level
                maxLevel = 0

                if len(profileContourElev[profileID]) > 0:
                    postfix = str(profileID).zfill(numDigits)

                    # create table for profile's levels properties
                    singleProfilePropsTablePath = targetWorkspace+os.sep+prefixPropsTable+postfix
                    try:
                        arcpy.AddMessage("    creating properties table ...")
                        arcpy.management.CreateTable (targetWorkspace, prefixPropsTable+postfix)
                        #arcpy.AddField_management(profilePropsTablePath, profileIDfieldName, "LONG", "", "", "", "", "", "", "")
                        arcpy.management.AddField(singleProfilePropsTablePath, levelFieldName , "SHORT", "", "", "", "", "", "", "")
                        arcpy.management.AddField(singleProfilePropsTablePath, damHeightFieldName , "DOUBLE", "", "", "", "", "", "", "")
                        arcpy.management.AddField(singleProfilePropsTablePath, elevationFieldName, "DOUBLE", "", "", "", "", "", "", "")
                        arcpy.management.AddField(singleProfilePropsTablePath, maxDepthFieldName, "DOUBLE", "", "", "", "", "", "", "")
                        arcpy.management.AddField(singleProfilePropsTablePath, deepestPntElevFieldName, "DOUBLE", "", "", "", "", "", "", "")
                        arcpy.management.AddField(singleProfilePropsTablePath, floodedAreaFieldName, "DOUBLE", "", "", "", "", "", "", "")
                        arcpy.management.AddField(singleProfilePropsTablePath, meanDepthFieldName, "DOUBLE", "", "", "", "", "", "", "")
                        arcpy.management.AddField(singleProfilePropsTablePath, volumeFieldName, "DOUBLE", "", "", "", "", "", "", "")
                        arcpy.management.AddField(singleProfilePropsTablePath, damLengthFieldName, "DOUBLE", "", "", "", "", "", "", "")
                        arcpy.management.AddField(singleProfilePropsTablePath, damVolumeFieldName, "DOUBLE", "", "", "", "", "", "", "")
                        arcpy.management.AddField(singleProfilePropsTablePath, damFVolumeFieldName, "DOUBLE", "", "", "", "", "", "", "")
                        arcpy.management.AddField(singleProfilePropsTablePath, volumeCoefFieldName, "DOUBLE", "", "", "", "", "", "", "")

                    except Exception as e:
                        arcpy.AddMessage(e)
                        arcpy.AddWarning("Failed creating profile properties table '"+singleProfilePropsTablePath+"'")            
                    else:
                        arcpy.AddMessage("    ... successfull")

                    # vybrat z hrazi pro nasledny prunik s vrstevnicema
                    where = arcpy.AddFieldDelimiters("profilesLayer", profileIDfieldName)+" = "+str(profileID)
                    arcpy.management.SelectLayerByAttribute ("profilesLayer", "NEW_SELECTION", where)
                    # vybrat z vnitrnich bodu pro nasledny prunik s polygonama
                    where = arcpy.AddFieldDelimiters("insidePointsLayer", profileIDfieldName)+" = "+str(profileID)
                    arcpy.management.SelectLayerByAttribute ("insidePointsLayer", "NEW_SELECTION", where)
                    # vybrat z charakteristických čar pro následný ořez a uložení
                    where = arcpy.AddFieldDelimiters("charLinesLayer", profileIDfieldName)+" = "+str(profileID)
                    arcpy.management.SelectLayerByAttribute ("charLinesLayer", "NEW_SELECTION", where)
                    # vybrat z průniku hladiny a hraze pro následné protnutí s vrstevnicí zátopy
                    where = arcpy.AddFieldDelimiters("waterLinesLayer", profileIDfieldName)+" = "+str(profileID)
                    arcpy.management.SelectLayerByAttribute ("waterLinesLayer", "NEW_SELECTION", where)

                    where = arcpy.AddFieldDelimiters("eraseLinesLayer", profileIDfieldName)+" = "+str(profileID)
                    arcpy.management.SelectLayerByAttribute ("eraseLinesLayer", "NEW_SELECTION", where)

                    elevList = profileContourElev[profileID]
                    #arcpy.AddMessage(str(elevList))

                    # profile's original clipped DEM
                    DEMfilePath = targetWorkspace+os.sep+DEMfileNamePrefix+postfix
                    arcpy.env.extent = arcpy.Describe(DEMfilePath).extent

                    numDigits2 = max(len(str(len(elevList))), 2)
                    e = 1
                    damHeight = 0

                    for curElev in elevList:
                        arcpy.AddMessage("    level {:d} ({:0.2f} m a.s.l.)".format(e, curElev))
                        damHeight += depthStep

                        # prepare the datasets names
                        postfix2 = "_e" + str(e).zfill(numDigits2)
                        DEMwDamFileName = DEMwDamFileNamePrefix+postfix+postfix2
                        DEMwDamFilePath = targetWorkspace+os.sep+DEMwDamFileName
                        damBodyDEMFileName = "damBodyDEM_"+postfix+postfix2
                        damBodyDEMFilePath = targetWorkspace+os.sep+damBodyDEMFileName

                        # create dam body DEM with absolute heights
                        damDEM = arcpy.ddd.Plus(targetWorkspace+os.sep+"damRaster_"+postfix, curElev, targetWorkspace+os.sep+"damRasterAbs_"+postfix+postfix2)
                        toBeDeleted.append(targetWorkspace+os.sep+"damRasterAbs_"+postfix+postfix2)

                        damRas = arcpy.Raster(damDEM)
                        DEMras = arcpy.Raster(DEMfilePath)
                        diffRas = damRas - DEMras 

                        # vytvoreni DEM se zahrnutou hrazi
                        DEMwDam = arcpy.sa.Con(arcpy.sa.IsNull(diffRas), DEMras, arcpy.sa.Con(diffRas, DEMras, damRas, "VALUE < 0"))
                        DEMwDam.save(DEMwDamFilePath)

                        # vytvoreni DEM "nadzemni casti hraze"
                        damBodyDEM = arcpy.sa.SetNull(diffRas, diffRas, "VALUE < 0")
                        damBodyDEM.save(damBodyDEMFileName)

                        # vytvoreni vrstevnice hladiny 
                        singleContourFilePath = targetWorkspace+os.sep+prefixCont+postfix+postfix2
                        arcpy.sa.ContourList (DEMfilePath, singleContourFilePath, [curElev])
                        toBeDeleted.append(singleContourFilePath)

                        # udelat jmeno datasetu
                        contPolyFileName = prefixContPoly+postfix+postfix2
                        contPolyFilePath = targetWorkspace+os.sep+contPolyFileName
                        # udelat polygon zatopene plochy prunikem s carou hladiny na hrazi
                        arcpy.management.FeatureToPolygon([singleContourFilePath, "waterLinesLayer"], contPolyFilePath, "", "ATTRIBUTES")
                        toBeDeleted.append(contPolyFilePath)

                        desc = arcpy.Describe(contPolyFilePath)
                        areaFieldName = desc.AreaFieldName 

                        # celkovy pocet vytvorenych polygonu pro danou uroven hladiny
                        totNum = int(arcpy.management.GetCount(contPolyFilePath).getOutput(0))
                        #arcpy.AddMessage("totNum = "+str(totNum))

                        tempPolyFilePath = targetWorkspace+os.sep+"tempFloodedArea_"+postfix+postfix2
                        toBeDeleted.append(tempPolyFilePath)

                        # pokud nebyl vytvoren ani jeden polygon
                        if totNum < 1:
                            minLevel = e + 1
                            arcpy.AddWarning("       No flooded area polygon created.")
                            props.append([e, damHeight, curElev, 0, 0, 0, 0, 0, 0, 0, 0, 0])

                        # pokud bylo neco vytvoreno ...
                        else:
                            # vybrat pouze ty, co jsou vevnitr nadrze
                            arcpy.management.MakeFeatureLayer (contPolyFilePath, "contourPolygonsLayers")
                            arcpy.management.SelectLayerByLocation("contourPolygonsLayers", "INTERSECT", "insidePointsLayer" , "", "NEW_SELECTION")

                            # spocitat vybrane (mel by byt jeden ...)
                            numSel = int(arcpy.management.GetCount("contourPolygonsLayers").getOutput(0))
                            #arcpy.AddMessage("numSel = "+str(numSel))

                            # pokud neni vybran zadny
                            if numSel == 0:
                                minLevel = e + 1
                                #profileTable.insertRow([e, damHeight, curElev, 0, 0, 0, 0, 0, 0, 0, 0])
                                props.append([e, damHeight, curElev, 0, 0, 0, 0, 0, 0, 0, 0, 0])
                                arcpy.AddWarning("       No flooded area polygon intersects with the inside point.")

                            # pokud je vybran vic nez jeden
                            elif numSel > 1:
                                arcpy.AddWarning("       More than one polygon selected by inside point ... this really shoudn't happen ...")
                                arcpy.AddWarning("       The rest of the calculation is definitely wrong!")

                            # pokud je vybran jeden (optimalni pripad)
                            else:
                                if e > maxLevel:
                                    maxLevel = e

                                # zkopirovat vybrany polygon zatopy
                                poly = arcpy.management.CopyFeatures ("contourPolygonsLayers", tempPolyFilePath)
                                arcpy.management.EliminatePolygonPart ("contourPolygonsLayers", tempPolyFilePath, "AREA", areaThreshold, "", "ANY")

                                # vybrat sousedni polygony po distStep krocich az do distLimit vzdalenosti
                                distLimit = 5*cellSize
                                distStep = 2*cellSize

                                while distLimit > 0:
                                    #arcpy.AddMessage(str(distLimit))
                                    arcpy.management.SelectLayerByLocation("contourPolygonsLayers", "INTERSECT", "contourPolygonsLayers" , str(distStep), "NEW_SELECTION")
                                    distLimit -= distStep

                                arcpy.management.SelectLayerByLocation("contourPolygonsLayers", "INTERSECT", poly , "", "REMOVE_FROM_SELECTION")

                                where = arcpy.AddFieldDelimiters("contourPolygonsLayers", areaFieldName) + " < "+str(areaThreshold)
                                #arcpy.AddMessage(where)
                                arcpy.management.SelectLayerByAttribute ("contourPolygonsLayers", "REMOVE_FROM_SELECTION", where)
                                arcpy.management.SelectLayerByLocation("contourPolygonsLayers", "INTERSECT", "eraseLinesLayer" , "", "REMOVE_FROM_SELECTION")

                                arcpy.management.Append(["contourPolygonsLayers"], poly, "NO_TEST")

                                polyFileName = prefixPoly+postfix+postfix2
                                polyFilePath = targetWorkspace+os.sep+polyFileName

                                arcpy.management.Dissolve (poly, polyFilePath, "", "", "MULTI_PART")

                                try:
                                    arcpy.management.AddField(polyFilePath, profileIDfieldName, "SHORT", "", "", "", "", "", "", "")
                                    arcpy.management.AddField(polyFilePath, damHeightFieldName , "DOUBLE", "", "", "", "", "", "", "")
                                    arcpy.management.AddField(polyFilePath, maxDepthFieldName, "DOUBLE", "", "", "", "", "", "", "")
                                    arcpy.management.AddField(polyFilePath, deepestPntElevFieldName, "DOUBLE", "", "", "", "", "", "", "")
                                    arcpy.management.AddField(polyFilePath, elevationFieldName, "DOUBLE", "", "", "", "", "", "", "")
                                except Exception as ex:
                                    arcpy.AddMessage(ex)
                                    arcpy.AddMessage("Failed creating attribute field in '"+polyFilePath+"'")
                                arcpy.management.CalculateField (polyFilePath, profileIDfieldName, str(profileID), "PYTHON_9.3")

                                # do the zonal statistics of the flooded area volume
                                statsTableFileName = prefixStatsTable+postfix+postfix2
                                statsTableFilePath = targetWorkspace+os.sep+statsTableFileName 

                                if arcpy.sa.ZonalStatisticsAsTable (polyFilePath, profileIDfieldName, DEMwDamFilePath, statsTableFilePath, "DATA", "ALL"):
                                    toBeDeleted.append(statsTableFilePath)

                                    with arcpy.da.SearchCursor(statsTableFilePath, ["AREA", "MEAN", "MIN"]) as table:
                                        for row in table:
                                            area = row[0]
                                            meanDepth = curElev - row[1]
                                            volume = area * meanDepth
                                            maxDepth = curElev - row[2]
                                            deepestPntElev = row[2]
                                else:
                                    arcpy.AddWarning("Error calculating flooded volume zonal stats for level "+str(level)+" (elevation "+str(curElev)+")!")

                                damStatsTableFileName = prefixStatsTable+postfix+postfix2
                                damStatsTableFilePath = targetWorkspace+os.sep+damStatsTableFileName

                                remap = arcpy.sa.RemapRange([[0, 10000, profileID]])
                                damPlanRaster = arcpy.sa.Reclassify (damBodyDEMFilePath, "VALUE", remap, "NODATA")
                                toBeDeleted.append(damPlanRaster)

                                with arcpy.da.UpdateCursor(polyFilePath, [damHeightFieldName, maxDepthFieldName, deepestPntElevFieldName, elevationFieldName]) as table:
                                    for row in table:
                                        row = [damHeight, maxDepth, deepestPntElev, curElev]
                                        table.updateRow(row)

                                damPlanFilePath = targetWorkspace+os.sep+damPlanPrefix+postfix+postfix2
                                arcpy.conversion.RasterToPolygon(damPlanRaster, damPlanFilePath, "SIMPLIFY", "VALUE")

                                if arcpy.sa.ZonalStatisticsAsTable (damPlanRaster, "VALUE", damBodyDEM, damStatsTableFilePath, "DATA", "ALL"):
                                    with arcpy.da.SearchCursor(damStatsTableFilePath, ["AREA", "SUM"]) as table:
                                        for row in table:
                                            damArea = row[0]
                                            damVolume = row[1] * cellSize * cellSize

                                    desc = arcpy.Describe(damPlanFilePath)
                                    perimeterFieldName = desc.LengthFieldName
                                    with arcpy.da.SearchCursor(damPlanFilePath, [perimeterFieldName]) as table:
                                        for row in table:
                                            foundationPerimeter = row[0]
                                            foundationArea = damArea-foundationPerimeter*damFoundationDepth

                                    damFoundationVolume = (damArea + foundationArea)/2*damFoundationDepth
                                    damVolumeF = damVolume+damFoundationVolume
                                else:
                                    arcpy.AddWarning("Error calculating dam volume zonal stats for level "+str(level)+" (elevation "+str()+")!")

                                volCoef = volume/damVolumeF

                                chLinesFilePath = targetWorkspace+os.sep+charLinesFileName+postfix+postfix2
                                arcpy.analysis.Clip("charLinesLayer", damPlanFilePath, chLinesFilePath)
                                if int(arcpy.management.GetCount(chLinesFilePath).getOutput(0)) > 2:
                                    arcpy.AddWarning("There seem to be too many dam body lines for level "+str(e)+" of profile #"+str(profileID)+".")
                                    arcpy.AddWarning("This is caused by insufficient dam model height ("+str(damModelHeight)+"m) in combination with high steepness of longitudal profile.")
                                #arcpy.AddMessage(str(profileProps))

                                # clip the dam line to get the length of dam for each water level
                                damAxisFilePath = targetWorkspace+os.sep+damAxisPrefix+postfix+postfix2

                                arcpy.analysis.Clip("profilesLayer", damPlanFilePath, damAxisFilePath)
                                # all except the last one can be deleted
                                # the last one can be usefull for maps ...
                                if e < len(elevList):
                                    toBeDeleted.append(damAxisFilePath)

                                # find the length of the dam midline
                                damLength = 0
                                with arcpy.da.SearchCursor(damAxisFilePath, ["Shape_Length"]) as damTable:
                                    for row in damTable:
                                        damLength = row[0]

                                # insert row into the results table with the calculated values
                                props.append([e, damHeight, curElev, maxDepth, deepestPntElev, area, meanDepth, volume, damLength, damVolume, damVolumeF, volCoef])

                        e += 1
                    with arcpy.da.InsertCursor(singleProfilePropsTablePath, [levelFieldName, damHeightFieldName , elevationFieldName, maxDepthFieldName, deepestPntElevFieldName,
                                                                             floodedAreaFieldName, meanDepthFieldName, volumeFieldName, damLengthFieldName, damVolumeFieldName, damFVolumeFieldName, volumeCoefFieldName]) as profileTable:
                        for row in props:
                            profileTable.insertRow(row)
                else:
                    minLevel = 0

            maxLevels.update({profileID: maxLevel})
            minLevels.update({profileID: minLevel})
       
        with arcpy.da.UpdateCursor(targetWorkspace+os.sep+profilesPropsTableName, [profileIDfieldName, minLevelFieldName, maxLevelFieldName]) as table:
            for row in table:
                if minLevels[row[0]] > maxLevels[row[0]]:
                    row[1] = None
                    row[2] = None
                else:
                    row[1] = minLevels[row[0]]
                    row[2] = maxLevels[row[0]]
                table.updateRow(row)
        
        arcpy.AddMessage("    ... successfull. ("+str(datetime.datetime.now()-startTime)+")")

 # FINAL CLEAN-UP -------------------------------------------------------------------------------------------------------
        if eraseIntermediateData:
            deleteTemporaryFiles(toBeDeleted)
            try:
                arcpy.AddMessage("Deleting TIN directory ...")
                os.rmdir(tindir)

            except Exception as e:
                arcpy.AddMessage(e)
                arcpy.AddWarning("Failed deleting TIN directory '"+tindir+"'")
            else:
                arcpy.AddMessage(" ... successful.")
        else:
            arcpy.AddWarning("All intermediate datasets were kept in the task geodatabase.")
        
        return

class FindConflicts(object):

    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "2. Find conflicts with infrastructure"
        self.description = "Finds conflicts of flooded areas with urban areas and linear infrastructure."
        self.canRunInBackground = False
        self.category = ""

    def getParameterInfo(self):
        """Define parameter definitions"""
        
        taskWorkspace = arcpy.Parameter(
            displayName = "Task geodatabase",
            name = "taskWorkspace",
            datatype = "DEWorkspace",
            parameterType = "Required",
            direction = "Input"
            )
            
        urbanAreas = arcpy.Parameter(
            displayName = "Urban areas definition",
            name = "urbanAreas",
            datatype = "GPFeatureLayer",
            parameterType = "Optional",
            direction = "Input"
            )
        urbanAreas.filter.list = ["Polygon"] 
        
        inputRoads = arcpy.Parameter(
            displayName = "Road network",
            name = "inputRoads",
            datatype = "GPFeatureLayer",
            parameterType = "Optional",
            direction = "Input"
            )
        inputRoads.filter.list = ["Polyline"]

        inputServiceRoads = arcpy.Parameter(
            displayName = "Service roads",
            name = "inputServiceRoads",
            datatype = "GPFeatureLayer",
            parameterType = "Optional",
            direction = "Input"
            )
        inputServiceRoads.filter.list = ["Polyline"]

        inputRailroads = arcpy.Parameter(
            displayName = "Railroad network",
            name = "inputRailroads",
            datatype = "GPFeatureLayer",
            parameterType = "Optional",
            direction = "Input"
            )
        inputRailroads.filter.list = ["Polyline"]

        params = [taskWorkspace, urbanAreas, inputRoads, inputServiceRoads, inputRailroads]  
        return params

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal validation is performed.  This method is called whenever a parameter has been changed.#"""
      
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool parameter.  This method is called after internal validation.#"""

        return

    def execute(self, parameters, messages):
        """The source code of the tool."""
        taskWorkspace = parameters[0].valueAsText
        urbanAreas = parameters[1].valueAsText
        inputRoads = parameters[2].valueAsText
        inputServiceRoads = parameters[3].valueAsText
        inputRailroads = parameters[4].valueAsText

        toBeDeleted = []
        
        profilesPropsTablePath = taskWorkspace+os.sep+profilesPropsTableName
#        desc = arcpy.Describe(profilesPropsTablePath)
#        nonIDfields = [minElevFieldName, maxLeftFieldName, maxRightFieldName, rangeFieldName, maxLevelFieldName, startElevFieldName]
#        nonIDfields.append(desc.OIDFieldName)        
        
        taskSettingsFilePath = taskWorkspace+os.sep+taskSettingsFileName
        profileIDfieldName = ""
        with arcpy.da.SearchCursor(taskSettingsFilePath, [IDfieldName]) as table:
            for row in table:
                profileIDfieldName = row[0]
        
        profileLevels = {}
        profileMins = {}
        
        fields = [profileIDfieldName, maxLevelFieldName]
        if startElevFieldName in getFieldNames(profilesPropsTablePath):
            fields.append(startElevFieldName)
        else:
            fields.append(minElevFieldName)
        
        with arcpy.da.SearchCursor(profilesPropsTablePath, fields) as table:
            for row in table:
                profileLevels.update({row[0]: row[1]})
                profileMins.update({row[0]: row[2]})        
        
        numDigits = max(len(str(max(profileLevels.keys()))), 2)
        # profileCollisions = {profileID: [urbanAreas, roads, railroads]}
        profileCollisions = {}
        
        for profileID in profileLevels.keys():
            arcpy.AddMessage("Processing profile #"+str(profileID)+" ...")
            maxLevel = int(profileLevels[profileID])
            postfix = str(profileID).zfill(numDigits)
            
            profilePropsTablePath = taskWorkspace+os.sep+prefixPropsTable+postfix        
            #fields = [damHeightFieldName ]
            fields = ["level"]
            
            if urbanAreas not in (None, "", " "):
                fields.append(urbanConfFieldName)
                if urbanConfFieldName not in getFieldNames(profilePropsTablePath):
                    try:
                        arcpy.management.AddField(profilePropsTablePath, urbanConfFieldName, "DOUBLE", "", "", "", "", "", "", "")
                    except Exception as e:
                        arcpy.AddMessage(e)
                        arcpy.AddMessage("Failed creating attribute field '"+urbanConfFieldName+"' in '"+profilePropsTablePath+"'")
            
            if inputRoads not in (None, "", " "):
                fields.append(roadsConfFieldName)
                if roadsConfFieldName not in getFieldNames(profilePropsTablePath):
                    try:
                        arcpy.management.AddField(profilePropsTablePath, roadsConfFieldName, "DOUBLE", "", "", "", "", "", "", "")
                    except Exception as e:
                        arcpy.AddMessage(e)
                        arcpy.AddMessage("Failed creating attribute field '"+roadsConfFieldName+"' in '"+profilePropsTablePath+"'")
            
            if inputServiceRoads not in (None, "", " "):
                fields.append(sRoadsConfFieldName)
                if sRoadsConfFieldName not in getFieldNames(profilePropsTablePath):
                    try:
                        arcpy.management.AddField(profilePropsTablePath, sRoadsConfFieldName, "DOUBLE", "", "", "", "", "", "", "")
                    except Exception as e:
                        arcpy.AddMessage(e)
                        arcpy.AddMessage("Failed creating attribute field '"+sRoadsConfFieldName+"' in '"+profilePropsTablePath+"'")
            
            if inputRailroads not in (None, "", " "):
                fields.append(railConfFieldName)
                if railConfFieldName not in getFieldNames(profilePropsTablePath):
                    try:
                        arcpy.management.AddField(profilePropsTablePath, railConfFieldName, "DOUBLE", "", "", "", "", "", "", "")
                    except Exception as e:
                        arcpy.AddMessage(e)
                        arcpy.AddMessage("Failed creating attribute field '"+railConfFieldName+"' in '"+profilePropsTablePath+"'")
                
            numDigits2 = max(len(str(maxLevel)), 2)
            arcpy.AddMessage(str(fields))
            with arcpy.da.UpdateCursor(profilePropsTablePath, fields) as table:
                for row in table:
                    level = row[0]
                    row = [level]
                    postfix2 = "_e" + str(level).zfill(numDigits2)

                    polyFileName = prefixPoly+postfix+postfix2
                    polyFilePath = taskWorkspace+os.sep+polyFileName  

                    if urbanAreas is not None:
                        urbanFilePath = taskWorkspace+os.sep+"xurban_"+postfix+postfix2
                        urbanCliped = arcpy.analysis.Clip(urbanAreas, polyFilePath, urbanFilePath)
                        toBeDeleted.append(urbanFilePath)
                        
                        if int(arcpy.management.GetCount(urbanCliped).getOutput(0)) > 0:
                            areaFieldName = arcpy.Describe(urbanCliped).AreaFieldName
                            totArea = 0
                            with arcpy.da.SearchCursor(urbanCliped, [areaFieldName]) as conflicts:
                                for item in conflicts:
                                    totArea += item[0]
                            row.append(totArea)
                        else:
                            row.append(0)
                            #arcpy.AddWarning("Profile #"+str(profileID)+" at level "+str(level)+" has conflict with urban area.")
                        
                    if inputRoads is not None:
                        roadsFilePath = taskWorkspace+os.sep+"xroads_"+postfix+postfix2
                        roadsCliped = arcpy.analysis.Clip(inputRoads, polyFilePath, roadsFilePath)
                        toBeDeleted.append(roadsFilePath)
                        
                        if int(arcpy.management.GetCount(roadsCliped).getOutput(0)) > 0:
                            lengthFieldName = arcpy.Describe(roadsCliped).LengthFieldName
                            totLength = 0
                            with arcpy.da.SearchCursor(roadsCliped, [lengthFieldName]) as conflicts:
                                for item in conflicts:
                                    totLength += item[0]
                            row.append(totLength)
                        else:
                            row.append(0)
                    
                    if inputServiceRoads is not None:
                        serviceRoadsFilePath = taskWorkspace+os.sep+"xservice_"+postfix+postfix2
                        serviceCliped = arcpy.analysis.Clip(inputServiceRoads, polyFilePath, serviceRoadsFilePath)
                        toBeDeleted.append(serviceRoadsFilePath)
                        
                        if int(arcpy.management.GetCount(serviceCliped).getOutput(0)) > 0:
                            lengthFieldName = arcpy.Describe(serviceCliped).LengthFieldName
                            totLength = 0
                            with arcpy.da.SearchCursor(serviceCliped, [lengthFieldName]) as conflicts:
                                for item in conflicts:
                                    totLength += item[0]
                            row.append(totLength)
                        else:
                            row.append(0)
                    
                    if inputRailroads is not None:
                        railroadsFilePath = taskWorkspace+os.sep+"xrail_"+postfix+postfix2
                        railCliped = arcpy.analysis.Clip(inputRailroads, polyFilePath, railroadsFilePath)
                        toBeDeleted.append(railroadsFilePath)
                    
                        if int(arcpy.management.GetCount(railCliped).getOutput(0)) > 0:
                            lengthFieldName = arcpy.Describe(railCliped).LengthFieldName
                            totLength = 0
                            with arcpy.da.SearchCursor(railCliped, [lengthFieldName]) as conflicts:
                                for item in conflicts:
                                    totLength += item[0]
                            row.append(totLength)
                        else:
                            row.append(0)
                    #arcpy.AddMessage(str(row))
                    table.updateRow(row)
            arcpy.AddMessage(" ... successsfull.")
        return


class AddResultsToMap(object):

    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Add processing results to current map project"
        self.description = "Add created features as layers with applied symbology to current map project."
        self.canRunInBackground = False
        self.category = ""

    def getParameterInfo(self):
        """Define parameter definitions"""
        taskWorkspace = arcpy.Parameter(
            displayName = "Task geodatabase",
            name = "taskWorkspace",
            datatype = "DEWorkspace",
            parameterType = "Required",
            direction = "Input"
            )
        chosenProfiles = arcpy.Parameter(
            displayName = "Choose profiles to be added to map",
            name = "chosenProfiles",
            datatype = "GPString",
            parameterType = "Optional",
            direction = "Input",
            multiValue = True
            )
        chosenProfiles.parameterDependencies = [taskWorkspace.name]
        chosenProfiles.filter.list = []
        
        addFloodedAreas = arcpy.Parameter(
            displayName = "Add flooded areas into map",
            name = "addFloodedAreas",
            datatype = "GPBoolean",
            parameterType = "Optional",
            direction = "Input"
            )
        addFloodedAreas.value = True
        
        flAreasLyr = arcpy.Parameter(
            displayName = "Symbology for flooded areas features",
            name = "flAreasLyr",
            datatype = "DEFile",
            parameterType = "Required",
            direction = "Input"
            )
        flAreasLyr.filter.list = ["lyr"]
        flAreasLyr.value = sys.path[0]+os.sep+"flarea.lyr"
        
        addDamBodyPlan = arcpy.Parameter(
            displayName = "Add dam body extent",
            name = "addDamBodyPlan",
            datatype = "GPBoolean",
            parameterType = "Optional",
            direction = "Input"
            )
        addDamBodyPlan.value = True
        
        damBodyLyr = arcpy.Parameter(
            displayName = "Symbology for dam extent",
            name = "damBodyLyr",
            datatype = "DEFile",
            parameterType = "Required",
            direction = "Input"
            )
        damBodyLyr.filter.list = ["lyr"]
        damBodyLyr.value = sys.path[0]+os.sep+"damplan.lyr"
        
        addCrestLines = arcpy.Parameter(
            displayName = "Add dam crest lines",
            name = "addCrestLines",
            datatype = "GPBoolean",
            parameterType = "Optional",
            direction = "Input"
            )
        addCrestLines.value = True
        
        damCrestLyr = arcpy.Parameter(
            displayName = "Symbology for dam crest lines",
            name = "damCrestLyr",
            datatype = "DEFile",
            parameterType = "Required",
            direction = "Input"
            )
        damCrestLyr.filter.list = ["lyr"]
        damCrestLyr.value = sys.path[0]+os.sep+"charlines.lyr"
        
        addMidLine = arcpy.Parameter(
            displayName = "Add dam mid line",
            name = "addMidLine",
            datatype = "GPBoolean",
            parameterType = "Optional",
            direction = "Input"
            )
        addMidLine.value = True
        
        damMidlineLyr = arcpy.Parameter(
            displayName = "Symbology for dam mid-line",
            name = "damMidlineLyr",
            datatype = "DEFile",
            parameterType = "Required",
            direction = "Input"
            )
        damMidlineLyr.filter.list = ["lyr"]
        damMidlineLyr.value = sys.path[0]+os.sep+"midline.lyr"
        
        addResultTables = arcpy.Parameter(
            displayName = "Add tables with results",
            name = "addResultTables",
            datatype = "GPBoolean",
            parameterType = "Optional",
            direction = "Input"
            )
        addResultTables.value = True
        
        params = [taskWorkspace, chosenProfiles, addFloodedAreas, flAreasLyr, addDamBodyPlan, damBodyLyr, addCrestLines, damCrestLyr, addMidLine, damMidlineLyr, addResultTables] 
        return params

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal validation is performed.  This method is called whenever a parameter has been changed.#"""
        if not parameters[0].hasBeenValidated:
            if parameters[0].altered:
                if parameters[0].value:
                    taskSettingsFilePath = parameters[0].valueAsText+os.sep+taskSettingsFileName
                    profileIDfieldName = ""
                    with arcpy.da.SearchCursor(taskSettingsFilePath, [IDfieldName]) as table:
                        for row in table:
                            profileIDfieldName = row[0]

                    profileIDlist = []
                    profileIDlistStr = []

                    profilesPropsTablePath = parameters[0].valueAsText+os.sep+profilesPropsTableName
                    with arcpy.da.SearchCursor(profilesPropsTablePath, [profileIDfieldName]) as table:
                        for row in table:
                            profileIDlist.append(row[0])
                            profileIDlistStr.append(str(row[0]))
                    parameters[1].filter.list = profileIDlist
                    parameters[1].value = profileIDlistStr

        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool parameter.  This method is called after internal validation.#"""
        if not parameters[1].value:
             parameters[1].setErrorMessage('At least one profile must be selected!')
        else:
            parameters[1].clearMessage() 
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""
        taskWorkspace = parameters[0].valueAsText
        IDlist = parameters[1].valueAsText.split(os.pathsep)
        addFloodedAreas = parameters[2].value
        flAreasLyr = parameters[3].valueAsText
        addDamBodyPlan = parameters[4].value
        damBodyLyr = parameters[5].valueAsText
        addCrestLines = parameters[6].value
        damCrestLyr = parameters[7].valueAsText
        addMidLine = parameters[8].value
        damMidlineLyr = parameters[9].valueAsText
        addResultTables = parameters[10].value

# READ THE VALUES FROM SETTINGS TABLE ----------------------------------------------------------------------------------------
        taskSettingsFilePath = taskWorkspace+os.sep+taskSettingsFileName
        profileIDfieldName = ""
        with arcpy.da.SearchCursor(taskSettingsFilePath, [IDfieldName]) as table:
            for row in table:
                profileIDfieldName = row[0]
       
        profilesPropsTablePath = taskWorkspace+os.sep+profilesPropsTableName
        
        maxID = 0
        maxLevels = {}
        minLevels = {}
        fields = [profileIDfieldName, minLevelFieldName, maxLevelFieldName]
        with arcpy.da.SearchCursor(profilesPropsTablePath, fields) as table:
            for row in table:
                maxID = max(maxID, row[0])
                if str(row[0]) in IDlist:
                    minLevels.update({row[0]: row[1]})
                    maxLevels.update({row[0]: row[2]})

        numDigits = max(len(str(maxID)), 2)
        
# REFERENCE THE MAP PROJECT AND DATAFRAME -----------------------------------------------------------------------------------
        mxd = arcpy.mapping.MapDocument("CURRENT")
        # get the data frame
        df = arcpy.mapping.ListDataFrames(mxd,"*")[0]
        df.referenceScale = 5000
        

# ADD THE LAYERS TO MAP AND APPLY SYMBOLOGY -----------------------------------------------------------------------------------
        #arcpy.AddMessage(str(minLevels))
        #arcpy.AddMessage(str(maxLevels))
        for pid in minLevels.keys():
            minLevel = minLevels[pid]
            maxLevel = maxLevels[pid]
            numDigits2 = max(len(str(maxLevel)), 2)
            if maxLevel == 0 and minLevel == 0:
                arcpy.AddMessage("Nothing to add for profile #"+str(pid)+"")
            else:
                arcpy.AddMessage("Adding layers of profile #"+str(pid)+" ...")
                postfix = str(pid).zfill(numDigits)
                groupLayer = arcpy.mapping.Layer(sys.path[0]+os.sep+"profile_group.lyr")  
                arcpy.mapping.AddLayer(df, groupLayer, "BOTTOM")

                layers = arcpy.mapping.ListLayers(mxd)
                #currentGroupLayer
                for layer in layers:

                    if layer.isGroupLayer and layer.name == "empty profile group":
                        layer.name = "["+postfix+"]"
                        currentGroupLayer = layer


                if addFloodedAreas is True:
                    # add all the flooded areas polygons
                    arcpy.AddMessage("  flooded areas")

                    transp = 80
                    finalTransp = 50

                    transpStep = int((transp-finalTransp)/(maxLevel-minLevel+1))

                    for level in range(minLevel, maxLevel+1):
                    #for level in range(minLevel, minLevel+2):

                        postfix2 = "_e" + str(level).zfill(numDigits2)
                        filePath = taskWorkspace+os.sep+prefixPoly+postfix+postfix2
                        if arcpy.Exists(filePath):
                            newlayer = arcpy.mapping.Layer(filePath)
                            arcpy.management.ApplySymbologyFromLayer(newlayer, flAreasLyr)
                            newlayer.transparency = transp
                            if transp > finalTransp:
                                transp -= transpStep

                            #arcpy.mapping.AddLayer(df, newlayer, "BOTTOM")
                            arcpy.mapping.AddLayerToGroup(df, currentGroupLayer, newlayer, "BOTTOM")
                        elif level != 0:
                            arcpy.AddWarning("'"+filePath+"' doesn't exist!")

                # add the dam mid line to the map
                if addMidLine is True:
                    arcpy.AddMessage("  dam midline")
                    filePath = taskWorkspace+os.sep+profilesFileName
                    if arcpy.Exists(filePath):
                        newlayer = arcpy.mapping.Layer(filePath)
                        #sourceLayer = arcpy.mapping.Layer(damMidlineLyr)
                        #arcpy.mapping.UpdateLayer (df, newlayer, sourceLayer, False)
                        
                        arcpy.management.ApplySymbologyFromLayer(newlayer, damMidlineLyr)
                        newlayer.definitionQuery = arcpy.AddFieldDelimiters(filePath, profileIDfieldName) + " = " + str(pid)
                        newlayer.name = "dam mid-line"
                        newlayer.visible = False

                        #arcpy.mapping.AddLayer(df, newlayer, "BOTTOM")
                        arcpy.mapping.AddLayerToGroup(df, currentGroupLayer, newlayer, "BOTTOM")
                    else:
                        arcpy.AddWarning("'"+filePath+"' doesn't exist!")

                # for the highest level add the characteristical lines of the dam body
                if addCrestLines is True:
                    arcpy.AddMessage("  dam body lines")
                    postfix2 = "_e" + str(maxLevel).zfill(numDigits2)
                    filePath = taskWorkspace+os.sep+charLinesFileName+postfix+postfix2
                    if arcpy.Exists(filePath):
                        newlayer = arcpy.mapping.Layer(filePath)
                        arcpy.management.ApplySymbologyFromLayer(newlayer, damCrestLyr)
                        newlayer.name = "dam crest"
                        #arcpy.mapping.AddLayer(df, newlayer, "BOTTOM")
                        arcpy.mapping.AddLayerToGroup(df, currentGroupLayer, newlayer, "BOTTOM")
                    else:
                        arcpy.AddWarning("'"+filePath+"' doesn't exist!")

                # and the extent of the dam body
                if addDamBodyPlan is True:
                    arcpy.AddMessage("  dam plan")
                    filePath = taskWorkspace+os.sep+damPlanPrefix+postfix+postfix2
                    if arcpy.Exists(filePath):
                        newlayer = arcpy.mapping.Layer(filePath)
                        arcpy.management.ApplySymbologyFromLayer(newlayer, damBodyLyr)
                        newlayer.name = "dam extent"
                        newlayer.transparency = 30
                        #arcpy.mapping.AddLayer(df, newlayer, "BOTTOM")
                        arcpy.mapping.AddLayerToGroup(df, currentGroupLayer, newlayer, "BOTTOM")
                    else:
                        arcpy.AddWarning("'"+filePath+"' doesn't exist!")
                        
                if addResultTables is True:
                    arcpy.AddMessage("  results table")
                    filePath = taskWorkspace+os.sep+prefixPropsTable+postfix
                    if arcpy.Exists(filePath):
                        tableView = arcpy.mapping.TableView(filePath)  
                        arcpy.mapping.AddTableView(df, tableView)
                    else:
                        arcpy.AddWarning("'"+filePath+"' doesn't exist!")
                arcpy.AddMessage("  ... successfull.")
        arcpy.RefreshTOC()
        arcpy.RefreshActiveView()
        return                
    
class CopyProperties(object):

    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Copy properties"
        self.description = "Disolves polygons of flooded areas and joins calculated properties to them."
        self.canRunInBackground = False
        self.category = ""

    def getParameterInfo(self):
        """Define parameter definitions"""
        
        inputWorkspace = arcpy.Parameter(
            displayName = "Input workspace",
            name = "inputWorkspace",
            datatype = "DEWorkspace",
            parameterType = "Required",
            direction = "Input"
            )
        outputWorkspace = arcpy.Parameter(
            displayName = "Output workspace",
            name = "outputWorkspace",
            datatype = "DEWorkspace",
            parameterType = "Required",
            direction = "Input"
            )
            
        params = [inputWorkspace, outputWorkspace]  
        return params

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal validation is performed.  This method is called whenever a parameter has been changed.#"""
      
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool parameter.  This method is called after internal validation.#"""

        return

    def execute(self, parameters, messages):
        """The source code of the tool."""
        inputWorkspace = parameters[0].valueAsText
        outputWorkspace = parameters[1].valueAsText
        
        profilesPropsTablePath = inputWorkspace+os.sep+profilesPropsTableName
#        desc = arcpy.Describe(profilesPropsTablePath)
#        nonIDfields = [minElevFieldName, maxLeftFieldName, maxRightFieldName, rangeFieldName, maxLevelFieldName, startElevFieldName]
#        nonIDfields.append(desc.OIDFieldName)        

        taskSettingsFilePath = inputWorkspace+os.sep+taskSettingsFileName
        profileIDfieldName = ""
        with arcpy.da.SearchCursor(taskSettingsFilePath, [IDfieldName]) as table:
            for row in table:
                profileIDfieldName = row[0]
        
        profileLevels = {}
        with arcpy.da.SearchCursor(profilesPropsTablePath, [profileIDfieldName, maxLevelFieldName]) as table:
            for row in table:
                profileLevels.update({row[0]: row[1]})
        
        numDigits = max(len(str(max(profileLevels.keys()))), 2)
        
        for profileID in profileLevels.keys():
            arcpy.AddMessage("Processing profile #"+str(profileID)+" ...")
            maxLevel = int(profileLevels[profileID])
            postfix = str(profileID).zfill(numDigits)
            
            profilePropsTablePath = inputWorkspace+os.sep+prefixPropsTable+postfix
            
            numDigits2 = max(len(str(maxLevel)), 2)
            for level in range(1, maxLevel+1):
                postfix2 = "_e" + str(level).zfill(numDigits2)

                polyFileName = prefixPoly+postfix+postfix2
                inPolyFilePath = inputWorkspace+os.sep+polyFileName  
                outpPolyFilePath = outputWorkspace+os.sep+polyFileName
                
                arcpy.management.Dissolve (inPolyFilePath, outpPolyFilePath, [damHeightFieldName ], "", "MULTI_PART")
                joinFields = [elevationFieldName, maxDepthFieldName, floodedAreaFieldName, meanDepthFieldName, volumeFieldName, damLengthFieldName, damVolumeFieldName, damFVolumeFieldName, volumeCoefFieldNam, urbanConfFieldName, roadsConfFieldName, sRoadsConfFieldName, railConfFieldName]
                arcpy.management.JoinField (outpPolyFilePath, damHeightFieldName, profilePropsTablePath, damHeightFieldName, joinFields)

        return
    
class CreateSummary(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Create task summary"
        self.description = "Disolves polygons of flooded areas and joins calculated properties to them."
        self.canRunInBackground = False
        self.category = ""

    def getParameterInfo(self):
        """Define parameter definitions"""
        
        taskWorkspace = arcpy.Parameter(
            displayName = "Task geodatabase",
            name = "taskWorkspace",
            datatype = "DEWorkspace",
            parameterType = "Required",
            direction = "Input"
            )
        outputTable = arcpy.Parameter(
            displayName = "Output summary csv table",
            name = "outputTable",
            datatype = "DEFile",
            parameterType = "Required",
            direction = "Output"
            )
            
        params = [taskWorkspace, outputTable]  
        return params

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal validation is performed.  This method is called whenever a parameter has been changed.#"""
        if parameters[1].altered:
            parameters[1].value = checkFileExtension(parameters[1].valueAsText, "csv")          
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool parameter.  This method is called after internal validation.#"""

        return

    def execute(self, parameters, messages):
        """The source code of the tool."""
        taskWorkspace = parameters[0].valueAsText
        outputTable = parameters[1].valueAsText
        
        profilesPropsTablePath = taskWorkspace+os.sep+profilesPropsTableName
#        desc = arcpy.Describe(profilesPropsTablePath)
#        nonIDfields = [minElevFieldName, maxLeftFieldName, maxRightFieldName, rangeFieldName, maxLevelFieldName, startElevFieldName]
#        nonIDfields.append(desc.OIDFieldName)        

        columnsDescriptions = {IDfieldName:"ID profilu", levelFieldName:"výpočetní krok", damHeightFieldName:"výška hráze [m]", elevationFieldName:"nadm. výška hladiny normálního nadržení", maxDepthFieldName:"maximální hloubka [m]", floodedAreaFieldName:"zatopená plocha [m2]", meanDepthFieldName:"průměrná hloubka [m]", volumeFieldName:"zadržený objem [m3]", damLengthFieldName:"délka hráze [m]", damVolumeFieldName:"objem tělesa hráze nad terénem", damFVolumeFieldName:"objem tělesa hráze včetně založení", volumeCoefFieldName:"objemový koeficient", urbanConfFieldName:"konflikt s intravilánem [m2]", roadsConfFieldName:"konflikt se silnicemi [m]", sRoadsConfFieldName:"konflikt s cestami [m]", railConfFieldName:"konflikt se železničními tratěmi [m]"}

        columnsHeaders = {IDfieldName:"ID", levelFieldName:"level", damHeightFieldName:"h [m]", elevationFieldName:"E [m n.m.]", maxDepthFieldName:"Hmax [m]", floodedAreaFieldName:"A [m2]", meanDepthFieldName:"Hmean [m]", volumeFieldName:"V [m3]", damLengthFieldName:"l [m]", damVolumeFieldName:"Vh [m3]", damFVolumeFieldName:"Vhz [m3]", volumeCoefFieldName:"Cv", urbanConfFieldName:"Xi [m2]", roadsConfFieldName:"Xs [m]", sRoadsConfFieldName:"Xc [m]", railConfFieldName:"Xz [m]"}

        columnsFormats = {levelFieldName:"%.0f", damHeightFieldName:"%.1f", elevationFieldName:"%.2f", maxDepthFieldName:"%.1f", floodedAreaFieldName:"%.0f", meanDepthFieldName:"%.2f", volumeFieldName:"%.0f", damLengthFieldName:"%.1f", damVolumeFieldName:"%.0f", damFVolumeFieldName:"%.0f", volumeCoefFieldName:"%.2f", urbanConfFieldName:"%.0f", roadsConfFieldName:"%.0f", sRoadsConfFieldName:"%.0f", railConfFieldName:"%.0f"}
        
        columnsToWrite = [levelFieldName, damHeightFieldName, elevationFieldName, maxDepthFieldName, floodedAreaFieldName, meanDepthFieldName, volumeFieldName, damLengthFieldName, damVolumeFieldName, damFVolumeFieldName, volumeCoefFieldName, urbanConfFieldName, roadsConfFieldName, sRoadsConfFieldName, railConfFieldName]
        
        
        # get the profile ID field name
        taskSettingsFilePath = taskWorkspace+os.sep+taskSettingsFileName
        profileIDfieldName = ""
        with arcpy.da.SearchCursor(taskSettingsFilePath, [IDfieldName]) as table:
            for row in table:
                profileIDfieldName = row[0]
        
        # get the list of profile IDs of the task
        profilesPropsTablePath = taskWorkspace+os.sep+profilesPropsTableName
        IDsList = []
        with arcpy.da.SearchCursor(profilesPropsTablePath, [profileIDfieldName]) as table:
            for row in table:
                IDsList.append(row[0])
        IDsList.sort()
        numDigits = max(len(str(max(IDsList))), 2)
        
        # add the field names from the first table ... the rest should be the same
        existingFields = getFieldNames(taskWorkspace+os.sep+prefixPropsTable+str(IDsList[0]).zfill(numDigits))
        outputColumns = []
        for f in columnsToWrite:
            if f in existingFields:
                outputColumns.append(f)
#        
#        if urbanConfFieldName in existingFields and urbanConfFieldName not in columnsToWrite:
#            columnsToWrite.append(urbanConfFieldName)
#        if roadsConfFieldName in existingFields and roadsConfFieldName not in columnsToWrite:
#            columnsToWrite.append(roadsConfFieldName)
#        if sRoadsConfFieldName in existingFields and sRoadsConfFieldName not in columnsToWrite:
#            columnsToWrite.append(sRoadsConfFieldName)
#        if railConfFieldName in existingFields and railConfFieldName not in columnsToWrite:
#            columnsToWrite.append(railConfFieldName)

        file = open(outputTable, 'w')
        # write the column description
        tostore = columnsDescriptions[IDfieldName]+";"
        for column in outputColumns:
            tostore += columnsDescriptions[column] + "; "
        file.write(tostore+"\n")
        # write the column headers
        tostore = columnsHeaders[IDfieldName]+";"
        for column in outputColumns:
            tostore += columnsHeaders[column]+"; "
        file.write(tostore+"\n")
        
        for pid in IDsList:
            postfix = str(pid).zfill(numDigits)
            singleProfilePropsTablePath = taskWorkspace+os.sep+prefixPropsTable+postfix

            if arcpy.Exists(singleProfilePropsTablePath):
                with arcpy.da.SearchCursor(singleProfilePropsTablePath, outputColumns) as table:
                    for row in table:
                        tostore = str(pid)+";"
                        for i in range(len(row)):
                            tostore += locale.format_string(columnsFormats[outputColumns[i]], row[i])+";"
                        file.write(tostore+"\n")
                arcpy.AddMessage("Profil #"+postfix+" done.")
        file.close()

        return
    
class BreachDams(object):

    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Breach dams"
        self.description = ""
        self.canRunInBackground = False
        locale.setlocale(locale.LC_NUMERIC, locale.getlocale())

    def getParameterInfo(self):
        """Define parameter definitions"""

        inputDEM = arcpy.Parameter(
            displayName = "Input DEM raster layer",
            name = "inputDEM",
            datatype = "GPRasterLayer",
            parameterType = "Required",
            direction = "Input"
            )
        inputBreachlines = arcpy.Parameter(
            displayName = "Input breach lines",
            name = "inputBreachlines",
            datatype = "GPFeatureLayer",
            parameterType = "Required",
            direction = "Input"
            )
        inputBreachlines.filter.list = ["Line"]

        outputDEM = arcpy.Parameter(
            displayName = "Output DEM",
            name = "outputDEM",
            datatype = "DERasterDataset",
            parameterType = "Optional",
            direction = "Output"
            )
        fillNewDEM = arcpy.Parameter(
            displayName = "Fill newly created DEM",
            name = "fillNewDEM",
            datatype = "GPBoolean",
            parameterType = "Required",
            direction = "Input"
            )
        fillNewDEM.value = True
        
        outputSinks = arcpy.Parameter(
            displayName = "Output sinks raster",
            name = "outputSinks",
            datatype = "DERasterDataset",
            parameterType = "Optional",
            direction = "Output",
            category = "Additional outputs"
            )
        outputProfilesCSV = arcpy.Parameter(
            displayName = "Output profiles csv table",
            name = "outputProfilesCSV",
            datatype = "DEFile",
            parameterType = "Optional",
            direction = "Output",
            category = "Additional outputs"
            )
        minSteepness = arcpy.Parameter(
            displayName = "Minimum steepness to be maintained along breachlines (positive fraction value)",
            name = "minSteepness",
            datatype = "GPDouble",
            parameterType = "Optional",
            direction = "Input",
            category = "Settings")
        minSteepness.value = 0.01
            
        params = [inputDEM, inputBreachlines, outputDEM, fillNewDEM, outputSinks, outputProfilesCSV, minSteepness]  
        return params

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal validation is performed.  This method is called whenever a parameter has been changed.#"""
      
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool parameter.  This method is called after internal validation.#"""

        return

    def execute(self, parameters, messages):
        """The source code of the tool."""
        inputDEM = parameters[0].valueAsText
        inputBreachlines = parameters[1].valueAsText
        outputDEM = parameters[2].valueAsText
        fillNewDEM = parameters[3].value
        outputSinks = parameters[4].valueAsText
        outputProfilesCSV = parameters[5].valueAsText
        minSteepness = locale.atof(str(parameters[6].valueAsText))*-1
        
        toBeDeleted = []
        sw = arcpy.env.scratchWorkspace
        
        try:
            cellSize = float(arcpy.management.GetRasterProperties (inputDEM, "CELLSIZEX").getOutput(0))
        except:
            arcpy.AddError("Failed obtaining input DEM cell size!")
        arcpy.env.extent = arcpy.Describe(inputDEM).extent
        
        try:
            startTime = datetime.datetime.now()
            arcpy.AddMessage("Getting elevations of breach lines ...") 
            breachlinesZ = sw+os.sep+"breachlinesZ"
            #arcpy.sa.ExtractValuesToPoints (profilePointsFilePath, inputDEM, profilePointsElevFilePath, "INTERPOLATE", "VALUE_ONLY")
            arcpy.ddd.InterpolateShape(inputDEM, inputBreachlines, breachlinesZ, "", "", "LINEAR", "DENSIFY")
            toBeDeleted.append(breachlinesZ)
            
        except Exception as e:
            arcpy.AddError(e.message)
            arcpy.AddError("Getting elevations of breach lines failed!")  
        else:
            arcpy.AddMessage("    ... successful! ("+str(datetime.datetime.now()-startTime)+")")         

        desc = arcpy.Describe(inputBreachlines)
        #areaFieldName = desc.AreaFieldName
        lengthFieldName = desc.LengthFieldName
        OIDFieldName = desc.OIDFieldName

        profilesTable = {}
        
        try:
            startTime = datetime.datetime.now()
            arcpy.AddMessage("Generating new elevation points ...") 
            newPoints = []
            
            with arcpy.da.SearchCursor(breachlinesZ, [shapeFieldName, lengthFieldName, OIDFieldName]) as profiles:
                for row in profiles:
                    distancesT = []
                    OelevationT = []
                    NelevationT = []
                    
                    startpt = row[0].firstPoint
                    endpt = row[0].lastPoint

                    elevchange = endpt.Z - startpt.Z
                    if elevchange > 0:
                        arcpy.AddWarning("Feature OID "+str(row[2])+" has negative elevation difference between endpoints. Check direction of the feature.")
                        
                    length = row[1]
                    #arcpy.AddMessage("Feature length: {} and elevation difference: {}".format(length, elevchange))

                    elevdiff = elevchange/length
                    if elevdiff > (minSteepness):
                        arcpy.AddWarning("Minimum steepness can't be maintained for feature OID "+str(row[2]))

                    # Step through each part of the feature
                    for part in row[0]:
                        num = len(part)
                        count = 0
                        distance = 0.0
                        cumulativeDistance = 0.0
                        cumulativeGroupDistance = 0.0
                        currentGroup = []

                        for point in part:
                            # for the first point
                            if count == 0:
                                newPoints.append(point)
                                
                                lastPoint = point
                                lastUsedZ = point.Z

                                NelevationT.append(point.Z)
                            
                            else:
                                currentPoint = point

                                distance = pow(pow(currentPoint.X - lastPoint.X, 2) + pow(currentPoint.Y - lastPoint.Y, 2), 0.5)
                                cumulativeDistance += distance
                                
                                if currentPoint.Z > (lastUsedZ + cumulativeGroupDistance * minSteepness):
                                    cumulativeGroupDistance += distance
                                    currentGroup.append([cumulativeGroupDistance, point])
                                    
                                    if not currentGroup:
                                        lastUsedZ = lastPoint.Z
                                    
                                    if currentGroup and count == num-1:
                                        arcpy.AddWarning("Minimum steepness could not be maintained for last segment of feature OID "+str(row[2]))
                                else:
                                    if currentGroup:
                                        #arcpy.AddMessage(str(currentGroup))
                                        cumulativeGroupDistance += distance
                                        elevationDifference = (currentPoint.Z-lastUsedZ)/cumulativeGroupDistance

                                        for p in currentGroup:
                                            dist = p[0]
                                            newZ = lastUsedZ+dist*elevationDifference
                                            newpnt = p[1]
                                            newpnt.Z = newZ
                                            newPoints.append(newpnt)
                                            NelevationT.append(newZ)
                                        currentGroup = []
                                        cumulativeGroupDistance = 0.0

                                    newPoints.append(point)
                                    lastUsedZ = point.Z
                                    newZ = point.Z
                                    NelevationT.append(newZ)
                                    
                                lastPoint = point

                            distancesT.append(cumulativeDistance)
                            OelevationT.append(point.Z)
                            
                            
                            count += 1
                    profilesTable.update({row[2]:[distancesT, OelevationT, NelevationT]})
                    
        except Exception as e:
            arcpy.AddError(e.message)
            arcpy.AddError("Generating new elevation points failed!")  
        else:
            arcpy.AddMessage("    ... successful! ("+str(datetime.datetime.now()-startTime)+")")      
            

        if outputProfilesCSV is not None:
            file = open(outputProfilesCSV, 'w')
            for id in profilesTable.keys():
                file.write("Profile ID:;"+str(id)+"\n")
                file.write("distance;")
                for d in profilesTable[id][0]:
                    file.write(str(d)+";")
                file.write("\n")
                
                file.write("orig. elev.;")
                for d in profilesTable[id][1]:
                    file.write(str(d)+";")
                file.write("\n")
                
                file.write("new elev.;")
                for d in profilesTable[id][2]:
                    file.write(str(d)+";")
                    
                file.write("\n\n")
            file.close()
            
        try:
            startTime = datetime.datetime.now()
            arcpy.AddMessage("Writing new points ...") 
            newZpointsFileName = "Zpoints"
            newZpointsFilePath = sw+os.sep+newZpointsFileName
            arcpy.management.CreateFeatureclass (sw, "Zpoints", "POINT", "", "DISABLED", "ENABLED", sr)
            pointElevFieldName = "elev"
            arcpy.management.AddField(newZpointsFilePath, pointElevFieldName, "DOUBLE", "", "", "", "", "", "", "")
            toBeDeleted.append(newZpointsFilePath)

            with arcpy.da.InsertCursor(newZpointsFilePath, ["SHAPE@", pointElevFieldName])as pointsCursor:
                for pnt in newPoints:
                    pointsCursor.insertRow([pnt, pnt.Z])
                    
        except Exception as e:
            arcpy.AddMessage(e)
            arcpy.AddMessage("Writing new points failed!")
        else:
            arcpy.AddMessage("    ... successful! ("+str(datetime.datetime.now()-startTime)+")")  

        try:
            startTime = datetime.datetime.now()
            arcpy.AddMessage("Converting points to raster ...") 
            
            rasterPointsFileName = "raster_Zpoints"
            rasterPointsFilePath = sw+os.sep+rasterPointsFileName            
            arcpy.conversion.PointToRaster(newZpointsFilePath, pointElevFieldName, rasterPointsFilePath, "MEAN", "", cellSize)
            toBeDeleted.append(rasterPointsFilePath)
            
        except Exception as e:
            arcpy.AddMessage(e)
            arcpy.AddMessage("Converting points to raster failed!")
        else:
            arcpy.AddMessage("    ... successful! ("+str(datetime.datetime.now()-startTime)+")")  

        if outputDEM is not None or outputSinks is not None:
            try:
                startTime = datetime.datetime.now()
                newDEM = arcpy.sa.Con(arcpy.sa.IsNull(rasterPointsFilePath), inputDEM, rasterPointsFilePath)

                if outputDEM is not None:
                    arcpy.AddMessage("Creating new DEM raster ...") 
                    if fillNewDEM:
                        toBeDeleted.append(newDEM)
                        filledDEM = arcpy.sa.Fill (newDEM)
                        filledDEM.save(outputDEM)
                        del filledDEM
                    else:
                        newDEM.save(outputDEM)
                        del newDEM

                if outputSinks is not None:
                    arcpy.AddMessage("Creating sinks raster ...") 
                    DEMflowDir = arcpy.sa.FlowDirection (newDEM, "FORCE")
                    toBeDeleted.append(DEMflowDir)
                    sinksRaster = arcpy.sa.Sink(DEMflowDir)
                    sinksRaster.save(outputSinks)

            except Exception as e:
                arcpy.AddMessage(e)
                arcpy.AddMessage("Creating new DEM raster failed!")
            else:
                arcpy.AddMessage("    ... successful! ("+str(datetime.datetime.now()-startTime)+")")  

# FINAL CLEAN-UP -------------------------------------------------------------------------------------------------------
        deleteTemporaryFiles(toBeDeleted)
        
        return

class CalculateLevelVolumeCurve(object):

    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Calculate level-volume curve for selected profile"
        self.description = "Calculates the characteristic level-volume curve for selected profile and dam height."
        self.canRunInBackground = False
        self.category = ""
        self.profileLimits = {}

    def getParameterInfo(self):
        """Define parameter definitions"""
        taskWorkspace = arcpy.Parameter(
            displayName = "Task geodatabase",
            name = "taskWorkspace",
            datatype = "DEWorkspace",
            parameterType = "Required",
            direction = "Input"
            )
        chosenProfile = arcpy.Parameter(
            displayName = "Choose profile to be processed",
            name = "chosenProfile",
            datatype = "GPLong",
            parameterType = "Required",
            direction = "Input",
            multiValue = False
            )
        chosenProfile.filter.list = []
        
        chosenLevel = arcpy.Parameter(
            displayName = "Choose dam height to be processed",
            name = "chosenLevel",
            datatype = "GPLong",
            parameterType = "Required",
            direction = "Input",
            multiValue = False
            )
        chosenLevel.filter.list = []

        compStep = arcpy.Parameter(
            displayName = "Incremental step for the water level",
            name = "compStep",
            datatype = "GPDouble",
            parameterType = "Required",
            direction = "Input",
            multiValue = False
            )

        outputTable = arcpy.Parameter(
            displayName = "Output level-volume csv table",
            name = "outputTable",
            datatype = "DEFile",
            parameterType = "Required",
            direction = "Output"
        )
        params = [taskWorkspace, chosenProfile, chosenLevel, compStep, outputTable]
        return params

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal validation is performed.  This method is called whenever a parameter has been changed.#"""
        #if not parameters[0].hasBeenValidated and parameters[0].altered:
        if parameters[0].altered and not parameters[0].hasBeenValidated:
            # if the GDB has changed, clear the dam level list and value
            parameters[1].filter.list = list()
            parameters[1].value = None
            parameters[2].filter.list = list()
            parameters[2].value = None

        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool parameter.  This method is called after internal validation.#"""
        if parameters[0].altered and parameters[0].value not in ("", " ", "#", None):
            taskSettingsFilePath = parameters[0].valueAsText+os.sep+taskSettingsFileName
            if arcpy.Exists(taskSettingsFilePath):
                #parameters[0].clearMessage()
                profileIDfieldName = ""
                with arcpy.da.SearchCursor(taskSettingsFilePath, [IDfieldName]) as table:
                    for row in table:
                       profileIDfieldName = row[0]

                profilesPropsTablePath = parameters[0].valueAsText+os.sep+profilesPropsTableName
                if arcpy.Exists(profilesPropsTablePath):
                    #parameters[0].clearMessage()
                    with arcpy.da.SearchCursor(profilesPropsTablePath, [profileIDfieldName, minLevelFieldName, maxLevelFieldName]) as table:
                        for row in table:
                            self.profileLimits.update({row[0]:range(row[1], row[2]+1)})
                            #parameters[2].setWarningMessage(str(self.profileLimits))
                    parameters[1].filter.list = self.profileLimits.keys()
                else:
                    parameters[0].setErrorMessage("The geodatabase does not containt the profile properties tables. The profiles must be processed first.")

            else:
                parameters[0].setErrorMessage("The geodatabase does not containt the settings table. The profiles must be processed first.")
        else:
            # if the GDB has changed, clear the dam level list and value
            parameters[1].filter.list = list()
            parameters[1].value = None
            parameters[2].filter.list = list()
            parameters[2].value = None

        if parameters[1].altered and parameters[1].value:
            parameters[2].filter.list = self.profileLimits[parameters[1].value]
            #parameters[2].value = None
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""
        taskWorkspace = parameters[0].valueAsText
        chosenProfileID = parameters[1].value
        chosenLevelID = parameters[2].value
        compStep = parameters[3].value
        outputFile = parameters[4].valueAsText

        toBeDeleted = []

        levelProps = {}

        dam = DamBody(3, 0.5, 3, 2, 1, "xxx", "yyy")
        arcpy.AddMessage(dam.damDescribtion())

# READ THE VALUES FROM SETTINGS TABLE ----------------------------------------------------------------------------------------
        arcpy.AddMessage("Reading the profiles properties ...")
        taskSettingsFilePath = taskWorkspace+os.sep+taskSettingsFileName

        profileIDfieldName = ""
        with arcpy.da.SearchCursor(taskSettingsFilePath, [IDfieldName]) as table:
            for row in table:
                profileIDfieldName = row[0]
       
        profilesPropsTablePath = taskWorkspace+os.sep+profilesPropsTableName
        
        maxID = 0
        maxLevel = 0
        minLevel = 0
        deepestPntElev = 0
        maxLevelElev = 0

        fields = [profileIDfieldName, minLevelFieldName, maxLevelFieldName]
        with arcpy.da.SearchCursor(profilesPropsTablePath, fields) as table:
            for row in table:
                maxID = max(maxID, row[0])
                if str(row[0]) == chosenProfileID:
                    minLevel = row[1]
                    maxLevel = row[2]

        numDigits = max(len(str(maxID)), 2)
        numDigits2 = max(len(str(maxLevel)), 2)

        # filename postfix of the profile ID
        postfix = str(chosenProfileID).zfill(numDigits)
        # filename postfix of the dam height level
        postfix2 = "_e" + str(chosenLevelID).zfill(numDigits2)

        # get the needed profile's info
        profilePropsTablePath = taskWorkspace + os.sep + prefixPropsTable + postfix
        with arcpy.da.SearchCursor(profilePropsTablePath, [levelFieldName, elevationFieldName, deepestPntElevFieldName]) as table:
            for row in table:
                if row[0] == chosenLevelID:
                    deepestPntElev = row[2]
                    maxLevelElev = row[1]

        # arcpy.AddMessage("Max water level elevation:" + str(maxLevelElev))
        # arcpy.AddMessage("Deepest point elevation:" + str(deepestPntElev))
        arcpy.AddMessage(" ... successfull.")

# GET THE WATER LEVEL ELEVATIONS -----------------------------------------------------------------
        arcpy.AddMessage("Preparing the elevation list for contour interpolation ...")
        # list of elevations to calculate the level-volume curve
        contourList = []

        # get the list by incrementing the elevations by the computation step, starting one step above the deepest point
        elev = deepestPntElev + compStep

        while (elev < maxLevelElev):
            contourList.append(elev)
            elev = elev + compStep
        # and add the maximum water level
        contourList.append(maxLevelElev)
        #arcpy.AddMessage(str(contourList))

        numDigits3 = max(len(str(len(contourList))), 2)
        arcpy.AddMessage(" ... successfull.")

# CALCULATE VOLUMES OF THE FLOODED AREAS -------------------------------------------------------------
        arcpy.AddMessage("Interpolating the contours and calculating the storage volume properties ...")
        # the "zero" level values [water level elevation, max depth, flooded area, volume]
        levelProps.update({0: [deepestPntElev, 0, 0, 0]})

        # prepare the maximum flooded area filepath
        maxFloodedArea = taskWorkspace + os.sep + prefixPoly + postfix + postfix2
        # prepare the DEM filepath
        DEMfilePath = taskWorkspace + os.sep + DEMwDamFileNamePrefix + postfix + postfix2
        if arcpy.Exists(DEMfilePath):
            #
            i = 1
            for elev in contourList:
                # filename postfix of the water level
                postfix3 = "_wl" + str(i).zfill(numDigits3)

                # create the contour
                singleContourFilePath = taskWorkspace + os.sep + prefixCont + postfix + postfix2 + postfix3


                arcpy.sa.ContourList(DEMfilePath, singleContourFilePath, [elev])
                toBeDeleted.append(singleContourFilePath)

                # create layer of the water level for selections
                arcpy.management.MakeFeatureLayer(singleContourFilePath, "waterLevelLayer")
                toBeDeleted.append("waterLevelLayer")
                # select polygons inside the maximum flooded area
                arcpy.management.SelectLayerByLocation("waterLevelLayer", "INTERSECT", maxFloodedArea, "", "NEW_SELECTION")
                # create polygons from contour polylines
                contPolyFilePath = taskWorkspace + os.sep + prefixContPoly + postfix + postfix2 + postfix3
                arcpy.management.FeatureToPolygon("waterLevelLayer", contPolyFilePath)
                toBeDeleted.append(contPolyFilePath)
                # dissolve into single polygon
                singlePolyFilePath = taskWorkspace + os.sep + prefixPoly + postfix + postfix2 + postfix3
                arcpy.management.Dissolve(contPolyFilePath, singlePolyFilePath)
                toBeDeleted.append(singlePolyFilePath)

                # zonal stats within the polygon of the DEM with dam body included
                statsTableFileName = prefixStatsTable + postfix + postfix2 + postfix3
                statsTableFilePath = taskWorkspace + os.sep + statsTableFileName
                idFieldName = arcpy.Describe(singlePolyFilePath).OIDFieldName

                if arcpy.sa.ZonalStatisticsAsTable(singlePolyFilePath, idFieldName, DEMfilePath, statsTableFilePath, "DATA", "ALL"):
                    toBeDeleted.append(statsTableFilePath)

                    # read the zonal stats result and save the values to a dictionary
                    with arcpy.da.SearchCursor(statsTableFilePath, ["AREA", "MEAN", "MIN"]) as table:
                        for row in table:
                            area = row[0]
                            meanDepth = elev - row[1]
                            volume = area * meanDepth
                            maxDepth = elev - row[2]
                            deepestPntElev = row[2]
                    levelProps.update({i:[elev, area, maxDepth, volume]})

                else:
                    arcpy.AddWarning("Error calculating flooded volume zonal stats for level " + str(i) + " (elevation " + str(elev) + ")!")
                    levelProps.update({i: [elev, 0, 0, 0]})

                arcpy.AddMessage("    " + locale.format("%.3f", elev) + "m a.s.l. done.")
                i += 1
        else:
            arcpy.AddError("The DEM dataset '"+DEMfilePath+"' not found.")
            arcpy.AddError("Profiles need to be processed first and all the resulting datasets must be kept.")
            arcpy.AddError("This tool can't proceed.")
            sys.exit(0)

        arcpy.AddMessage(" ... successfull.")

# WRITE THE RESULTS TO FILES -------------------------------------------------------------
        arcpy.AddMessage("Writing the results to files ...")
        outputTableFileName = prefixVolumes + postfix + postfix2
        outputTableFilePath = os.path.join(taskWorkspace, outputTableFileName)
        try:

            arcpy.management.CreateTable(taskWorkspace, outputTableFileName)
            arcpy.management.AddField(outputTableFilePath, wLevelIDFieldName, "SHORT", "", "", "", "", "", "", "")
            arcpy.management.AddField(outputTableFilePath, elevationFieldName, "DOUBLE", "", "", "", "", "", "", "")
            arcpy.management.AddField(outputTableFilePath, maxDepthFieldName, "DOUBLE", "", "", "", "", "", "", "")
            arcpy.management.AddField(outputTableFilePath, floodedAreaFieldName, "DOUBLE", "", "", "", "", "", "", "")
            arcpy.management.AddField(outputTableFilePath, volumeFieldName, "DOUBLE", "", "", "", "", "", "", "")

        except Exception as e:
            arcpy.AddMessage(e)
            arcpy.AddMessage("    Failed creating output volumes table '" + outputTableFilePath + "'")
        else:
            arcpy.AddMessage("    Results table successfully created.")

        with arcpy.da.InsertCursor(outputTableFilePath, [wLevelIDFieldName, elevationFieldName, maxDepthFieldName, floodedAreaFieldName, volumeFieldName]) as table:
            for lwID in levelProps.keys():
                row = [lwID]
                row.extend(levelProps[lwID])
                table.insertRow(row)

        arcpy.AddMessage("    Results table successfully filled.")
        #arcpy.AddMessage(str(levelProps))

# WRITE RESULTS TO TABLE -------------------------------------------------------------------
        columnsDescriptions = {
            elevationFieldName: "nadm. výška hladiny",
            floodedAreaFieldName: "zatopená plocha [m2]",
            maxDepthFieldName: "hloubka [m]",
            volumeFieldName: "zadržený objem [m3]"}

        columnsHeaders = {
            elevationFieldName: "WL [m n.m.]",
            floodedAreaFieldName: "A [m2]",
            maxDepthFieldName: "h [m]",
            volumeFieldName: "V [m3]"}

        columnsFormats = {
            elevationFieldName: "%.2f",
            floodedAreaFieldName: "%.0f",
            maxDepthFieldName: "%.2f",
            volumeFieldName: "%.0f"}

        outputColumns = [elevationFieldName, floodedAreaFieldName, maxDepthFieldName, volumeFieldName]

        try:
            file = open(outputFile, 'w')
            # write the task info
            tostore = "ID profilu;"+str(chosenProfileID) + "\n"
            tostore += "Výška hráze;"+str(chosenLevelID) + "\n"
            file.write(tostore + "\n")

            # write the column description
            tostore = "ID výpočetního kroku;"
            for column in outputColumns:
                tostore += columnsDescriptions[column] + "; "
            file.write(tostore + "\n")

            # write the column headers
            tostore = ";"
            for column in outputColumns:
                tostore += columnsHeaders[column] + "; "
            file.write(tostore + "\n")

            for wlID in levelProps.keys():
                tostore = str(wlID) + ";"
                row = levelProps[wlID]
                for i in range(len(row)):
                    tostore += locale.format_string(columnsFormats[outputColumns[i]], row[i]) + ";"
                file.write(tostore + "\n")
            file.close()
        except Exception as e:
            arcpy.AddMessage(e)
            arcpy.AddMessage("    Failed creating output volumes CSV file '" + outputFile + "'")
        else:
            arcpy.AddMessage("    Results CSV file successfully created and filled.")

# FINAL CLEAN-UP -------------------------------------------------------------------------------------------------------
        #deleteTemporaryFiles(toBeDeleted)
        return


class GetDamsWatershed(object):

    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Create watersheds"
        self.description = "Create watersheds for each of the dams in task directory processed by the MVN tool"
        self.canRunInBackground = False
        self.category = ""

    def getParameterInfo(self):
        """Define parameter definitions"""
        taskWorkspace = arcpy.Parameter(
            displayName = "Task geodatabase",
            name = "taskWorkspace",
            datatype = "DEWorkspace",
            parameterType = "Required",
            direction = "Input"
            )

        createVectorWatersheds = arcpy.Parameter(
            displayName="Create polygon definition of watersheds",
            name="createVectorWatersheds",
            datatype="GPBoolean",
            parameterType="Optional",
            direction="Input",
        )
        createVectorWatersheds.value = True

        vectorPrefix = arcpy.Parameter(
            displayName="Vector watershed filename prefix",
            name="vectorPrefix",
            datatype="GPString",
            parameterType="Optional",
            direction="Input",
            category="Settings"
        )
        vectorPrefix.value = "ws_"

        preserveRasterWatersheds = arcpy.Parameter(
            displayName="Preserve raster definition of watersheds",
            name="preserveRasterWatersheds",
            datatype="GPBoolean",
            parameterType="Optional",
            direction="Input",
        )
        preserveRasterWatersheds.value = False

        rasterPrefix = arcpy.Parameter(
            displayName="Raster watershed filename prefix",
            name="rasterPrefix",
            datatype="GPString",
            parameterType="Optional",
            direction="Input",
            category="Settings"
        )
        rasterPrefix.value = "ras_ws_"

        writeAreaToResults = arcpy.Parameter(
            displayName="Write the watersheds' area to result table",
            name="writeAreaToResults",
            datatype="GPBoolean",
            parameterType="Optional",
            direction="Input",
        )
        params = [taskWorkspace, createVectorWatersheds, vectorPrefix, preserveRasterWatersheds, rasterPrefix, writeAreaToResults]
        return params

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal validation is performed.  This method is called whenever a parameter has been changed.#"""
        if not parameters[1].hasBeenValidated:
            if parameters[1].value == 1:
                parameters[2].enabled = 1
                parameters[2].parameterType = "Requiered"
            if parameters[1].value == 0:
                parameters[2].enabled = 0
                parameters[2].parameterType = "Optional"
        if parameters[1].altered:
            if parameters[1].value:
                if parameters[1].value == 1:
                    parameters[2].enabled = 1
                    parameters[2].parameterType = "Requiered"
                if parameters[1].value == 0:
                    parameters[2].enabled = 0
                    parameters[2].parameterType = "Optional"

        if not parameters[3].hasBeenValidated:
            if parameters[3].value == 1:
                parameters[4].enabled = 1
                parameters[4].parameterType = "Requiered"
            if parameters[3].value == 0:
                parameters[4].enabled = 0
                parameters[4].parameterType = "Optional"
        if parameters[3].altered:
            if parameters[3].value:
                if parameters[3].value == 1:
                    parameters[4].enabled = 1
                    parameters[4].parameterType = "Requiered"
                if parameters[3].value == 0:
                    parameters[4].enabled = 0
                    parameters[4].parameterType = "Optional"
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool parameter.  This method is called after internal validation.#"""

        return

    def execute(self, parameters, messages):
        """The source code of the tool."""
        taskWorkspace = parameters[0].valueAsText
        createVectorWatersheds = parameters[1].value
        vectorPrefix = parameters[2].valueAsText
        preserveRasterWatersheds = parameters[3].value
        rasterPrefix = parameters[4].valueAsText
        writeAreaToResults = parameters[5].value

        if not createVectorWatersheds and not preserveRasterWatersheds:
            arcpy.AddError("No raster watersheds ... no vector watersheds ... nothing to do here ...")
            return

# READ THE VALUES FROM SETTINGS TABLE ----------------------------------------------------------------------------------------
        taskSettingsFilePath = taskWorkspace + os.sep + taskSettingsFileName
        profileIDfieldName = ""
        with arcpy.da.SearchCursor(taskSettingsFilePath, [IDfieldName, taskDEMFieldName]) as table:
            for row in table:
                profileIDfieldName = row[0]
                inputDEM = row[1]

# READ THE VALUES FROM PROFILES PROPERTIES TABLE ----------------------------------------------------------------------------------------
        profilesPropsTablePath = taskWorkspace + os.sep + profilesPropsTableName
        # get maximum ID and level for file naming purposes
        maxID = 0
        # get dictionary of profile IDs and their maximum levels
        maxLevels = {}
        with arcpy.da.SearchCursor(profilesPropsTablePath, [profileIDfieldName, maxLevelFieldName]) as table:
            for row in table:
                maxLevels.update({row[0]:row[1]})
                maxID = max(maxID, row[0])

        numDigits = max(len(str(maxID)), 2)

        ext = ""
        desc = arcpy.Describe(taskWorkspace)
        if desc.dataType == "Folder":
            ext = ".shp"

# PREPARE NEEDED DEM DERIVATIVES ----------------------------------------------------------------------------------------------------------
        DEMfilledFileName = "DEM_filled"
        DEMflowDirFileName = "DEM_flowdir"

        toBeDeleted = []

        DEMfilledFilePath = taskWorkspace + os.sep + DEMfilledFileName
        # create filled DEM raster
        if not arcpy.Exists(DEMfilledFilePath):
            arcpy.AddMessage("Filling DEM ... ")
            DEMfilled = arcpy.sa.Fill(inputDEM)
            DEMfilled.save(DEMfilledFilePath)
            toBeDeleted.append(DEMfilledFilePath)
            arcpy.AddMessage(" ... successfull.")

        DEMflowDirFilePath = taskWorkspace + os.sep + DEMflowDirFileName
        # create flow direction raster
        if not arcpy.Exists(DEMflowDirFilePath):
            arcpy.AddMessage("Calculating flow direction raster ... ")
            DEMflowDir = arcpy.sa.FlowDirection(DEMfilledFilePath, "NORMAL")
            DEMflowDir.save(DEMflowDirFilePath)
            toBeDeleted.append(DEMflowDirFilePath)
            arcpy.AddMessage(" ... successfull.")

        # set the snap raster
        arcpy.env.snapRaster = inputDEM
        arcpy.env.extent = arcpy.Describe(inputDEM).extent
        # set the processing cell size
        CellSizeX = arcpy.management.GetRasterProperties(inputDEM, "CELLSIZEX").getOutput(0)
        CellSizeY = arcpy.management.GetRasterProperties(inputDEM, "CELLSIZEY").getOutput(0)
        if not CellSizeX == CellSizeY:
            arcpy.AddError("Reference raster X (" + str(CellSizeX) + ") and Y (" + str(CellSizeY) + ") cellsize is not equal")
        else:
            arcpy.env.cellSize = CellSizeX

# PROCESS WATERSHEDS ---------------------------------------------------------------------------------------------------------------
        j = 1
        areas = {}
        for id in maxLevels.keys():
            arcpy.AddMessage("Creating watershed definition for profile {:d} ({:d}/{:d})".format(id, j, len(maxLevels.keys())))
            postfix = str(id).zfill(numDigits)
            numDigits2 = max(len(str(maxLevels[id])), 2)
            # only the max level is processed
            postfix2 = "_e"+str(maxLevels[id]).zfill(numDigits2)

            # create name of the single export file with the selection
            profileRasterPath = taskWorkspace + os.sep + "profile_ras_" + postfix
            outWatershed_ras = taskWorkspace + os.sep + rasterPrefix + postfix
            outWatershed_vec = taskWorkspace + os.sep + vectorPrefix + postfix + ext

            arcpy.AddMessage("   converting profile to raster ...")
            damMidline = taskWorkspace+os.sep+damAxisPrefix+postfix+postfix2
            profileRaster = arcpy.conversion.FeatureToRaster (damMidline, profileIDfieldName, profileRasterPath, arcpy.env.cellSize)
            toBeDeleted.append(profileRaster)
            arcpy.AddMessage("   ... successfull")

            if arcpy.Exists(profileRasterPath) and not arcpy.Exists(outWatershed_ras):
                arcpy.AddMessage("   calculating raster watershed ...")
                # create the watershed
                ws = arcpy.sa.Watershed(DEMflowDirFilePath, profileRaster, "Value")
                ws.save(outWatershed_ras)
                if not preserveRasterWatersheds:
                    toBeDeleted.append(outWatershed_ras)
                arcpy.AddMessage("   ... successfull")

            if createVectorWatersheds:
                if arcpy.Exists(outWatershed_ras):
                    arcpy.AddMessage("   converting watershed to polygon ...")
                    # vectorize the watershed
                    arcpy.conversion.RasterToPolygon(outWatershed_ras, outWatershed_vec, "SIMPLIFY")
                    if profileIDfieldName not in getFieldNames(outWatershed_vec):
                        # ad profile id into resulting polygon layer
                        arcpy.management.AddField(outWatershed_vec, profileIDfieldName, "SHORT", "", "", "", "", "", "", "")

                    areaFN = arcpy.Describe(outWatershed_vec).AreaFieldName
                    with arcpy.da.UpdateCursor(outWatershed_vec, [profileIDfieldName, areaFN]) as table:
                        for row in table:
                            row[0] = id
                            areas.update({id:row[1]})
                            table.updateRow(row)
                    arcpy.AddMessage("   ... successfull")
            j += 1

# WRITE WATERSHED AERAS INTO PROFILES PROPERTIES TABLE ---------------------------------------------------------------------------------------------------
        if writeAreaToResults:
            arcpy.AddMessage("Writing results to profile properties table ...")
            if watershedAreaFieldName not in getFieldNames(profilesPropsTablePath):
                # ad profile id into resulting polygon layer
                arcpy.management.AddField(profilesPropsTablePath, watershedAreaFieldName, "DOUBLE", "", "", "", "", "", "", "")

            with arcpy.da.UpdateCursor(profilesPropsTablePath, [profileIDfieldName, watershedAreaFieldName]) as table:
                for row in table:
                    row[1] = areas[row[0]]
                    table.updateRow(row)
            arcpy.AddMessage("   ... successfull")

        # FINAL CLEAN-UP -------------------------------------------------------------------------------------------------------
        deleteTemporaryFiles(toBeDeleted)

        return
#
# class DamBody:
#     # how deep should be the dam TIN model be created? [m]
#     totalHight = 20
#
#     def __init__(self, crestWidth, saveHeight, waterwardSteepness, windwardSteepness, foundationDepth, DEMpath, midLine):
#         self.midLine = midLine
#         self.crestWidth = crestWidth
#         self.saveHeight = saveHeight
#         self.waterwardSteepness = waterwardSteepness
#         self.windwardSteepness = windwardSteepness
#         self.foundationDepth = foundationDepth
#         self.DEMpath = DEMpath
#
#     def damDescribtion(self):
#         return "Dam body with following properties: {}, 1:{}, 1:{}, {}, {}, {}, {}".format(self.crestWidth, self.waterwardSteepness, self.windwardSteepness, self.saveHeight, \
#                self.foundationDepth, self.midLine, self.DEMpath)
